from mage_ai.services.k8s.job_manager import JobManager
from mage_ai.tests.base_test import TestCase
from unittest.mock import MagicMock, call, patch
import os


@patch.dict(os.environ, {'HOSTNAME': 'mage-server'})
class JobManagerTests(TestCase):
    @patch('mage_ai.services.k8s.job_manager.config')
    def test_load_config(self, mock_config):
        with patch.object(mock_config, 'load_incluster_config') as mock_load_incluster_config:
            self.assertTrue(JobManager.load_config())
            mock_load_incluster_config.assert_called_once()

    @patch('mage_ai.services.k8s.job_manager.client')
    @patch('mage_ai.services.k8s.job_manager.config')
    def test_init(self, mock_config, mock_client):
        mock_batch_api_client = MagicMock()
        mock_core_v1_api_client = MagicMock()
        mock_pod_config = MagicMock()

        with patch.object(JobManager, 'load_config') as mock_load_config:
            with patch.object(mock_client, 'BatchV1Api', return_value=mock_batch_api_client):
                with patch.object(mock_client, 'CoreV1Api', return_value=mock_core_v1_api_client):
                    with patch.object(
                        mock_core_v1_api_client,
                        'read_namespaced_pod',
                        return_value=mock_pod_config,
                    ) as mock_read_namespaced_pod_method:

                        job_manager = JobManager(
                            job_name='test_job_name',
                            namespace='test_namespace',
                        )
                        self.assertEqual(job_manager.job_name, 'test_job_name')
                        self.assertEqual(job_manager.namespace, 'test_namespace')
                        self.assertEqual(job_manager.batch_api_client, mock_batch_api_client)
                        self.assertEqual(job_manager.api_version, 'batch/v1')
                        self.assertEqual(job_manager.core_api_client, mock_core_v1_api_client)
                        self.assertEqual(job_manager.pod_config, mock_pod_config)

                        mock_load_config.assert_called_once()
                        mock_read_namespaced_pod_method.assert_called_once_with(
                            name='mage-server',
                            namespace='test_namespace',
                        )

    @patch('mage_ai.services.k8s.job_manager.client')
    @patch('mage_ai.services.k8s.job_manager.config')
    def test_create_job_object(self, mock_config, mock_client):
        mock_v1_container = MagicMock()
        mock_v1_pod_template_spec = MagicMock()
        mock_v1_object_meta = MagicMock()
        mock_v1_pod_spec = MagicMock()
        mock_v1_job_spec = MagicMock()
        mock_v1_job = MagicMock()

        mock_client.V1Container = MagicMock(return_value=mock_v1_container)
        mock_client.V1PodTemplateSpec = MagicMock(return_value=mock_v1_pod_template_spec)
        mock_client.V1ObjectMeta = MagicMock(return_value=mock_v1_object_meta)
        mock_client.V1PodSpec = MagicMock(return_value=mock_v1_pod_spec)
        mock_client.V1JobSpec = MagicMock(return_value=mock_v1_job_spec)
        mock_client.V1Job = MagicMock(return_value=mock_v1_job)

        job_manager = JobManager(
            job_name='test_job_name',
            namespace='test_namespace',
        )
        command = 'mage run test_pipeline'
        job_manager.create_job_object(command)
        mock_client.V1Container.assert_called_once_with(
            name='mage-job-container',
            image=job_manager.pod_config.spec.containers[0].image,
            image_pull_policy='IfNotPresent',
            command=['mage', 'run', 'test_pipeline'],
            env=job_manager.pod_config.spec.containers[0].env,
            volume_mounts=job_manager.pod_config.spec.containers[0].volume_mounts,
        )
        mock_client.V1ObjectMeta.assert_has_calls(
            [
                call(labels={'name': 'test_job_name'}),
                call(name='test_job_name'),
            ]
        )
        mock_client.V1PodSpec.assert_called_once_with(
            restart_policy='Never',
            containers=[mock_v1_container],
            volumes=job_manager.pod_config.spec.volumes,
            image_pull_secrets=job_manager.pod_config.spec.image_pull_secrets,
        )
        mock_client.V1JobSpec.assert_called_once_with(
            template=mock_v1_pod_template_spec,
            backoff_limit=0,
        )
        mock_client.V1Job.assert_called_once_with(
            api_version='batch/v1',
            kind='Job',
            metadata=mock_v1_object_meta,
            spec=mock_v1_job_spec,
        )

    @patch('mage_ai.services.k8s.job_manager.client')
    @patch('mage_ai.services.k8s.job_manager.config')
    def test_create_job(self, mock_config, mock_client):
        mock_api_client = MagicMock()
        mock_api_client.create_namespaced_job = MagicMock()
        mock_job = MagicMock()

        mock_client.BatchV1Api = MagicMock(return_value=mock_api_client)
        job_manager = JobManager(
            job_name='test_job_name',
            namespace='test_namespace',
        )

        job_manager.create_job(mock_job)

        mock_api_client.create_namespaced_job.assert_called_once_with(
            body=mock_job,
            namespace='test_namespace',
        )

    @patch('mage_ai.services.k8s.job_manager.client')
    @patch('mage_ai.services.k8s.job_manager.config')
    def test_delete_job(self, mock_config, mock_client):
        mock_api_client = MagicMock()
        mock_api_client.delete_namespaced_job = MagicMock()
        mock_delete_options = MagicMock()
        mock_client.V1DeleteOptions = MagicMock(return_value=mock_delete_options)

        mock_client.BatchV1Api = MagicMock(return_value=mock_api_client)
        job_manager = JobManager(
            job_name='test_job_name',
            namespace='test_namespace',
        )

        job_manager.delete_job()

        mock_client.V1DeleteOptions.assert_called_once_with(
            propagation_policy='Foreground',
            grace_period_seconds=0,
        )
        mock_api_client.delete_namespaced_job.assert_called_once_with(
            name='test_job_name',
            namespace='test_namespace',
            body=mock_delete_options
        )



from mage_ai.settings import SERVER_VERBOSITY
from mage_ai.shared.logger import LoggingLevel
import logging


class Logger:
    def new_server_logger(self, name):
        return self.new_logger_with_verbosity(name, SERVER_VERBOSITY)

    def new_logger_with_verbosity(self, name, level: str):
        server_logger = logging.getLogger(name)
        if LoggingLevel.is_valid_level(level):
            server_logger.setLevel(level.upper())
        else:
            print(f"invalid verbosity {level} for logger {name}")
        return server_logger



"""Add completed_at.

Revision ID: 053ee2c10d85
Revises: 52ab80005742
Create Date: 2022-08-31 01:22:18.258615

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '053ee2c10d85'
down_revision = '52ab80005742'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('block_run', sa.Column('completed_at', sa.DateTime(timezone=True), nullable=True))
    op.add_column('pipeline_run', sa.Column('completed_at', sa.DateTime(timezone=True), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipeline_run', 'completed_at')
    op.drop_column('block_run', 'completed_at')
    # ### end Alembic commands ###



from mage_ai.streaming.sources.amazon_sqs import AmazonSqsSource
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class AmazonSqsTests(TestCase):
    def test_init(self):
        with patch.object(AmazonSqsSource, 'init_client') as mock_init_client:
            AmazonSqsSource(dict(
                connector_type='amazon_sqs',
                queue_name='test_queue',
                batch_size=1,
                wait_time_seconds=1,
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(AmazonSqsSource, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                AmazonSqsSource(dict(
                    connector_type='amazon_sqs',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'queue_name\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)

    def test_init_with_serde_config(self):
        with patch.object(AmazonSqsSource, 'init_client') as mock_init_client:
            source = AmazonSqsSource(dict(
                connector_type='kafka',
                queue_name='test_queue',
                serde_config=dict(
                    serialization_method='JSON',
                )
            ))
            mock_init_client.assert_called_once()
            self.assertEqual(source.config.serde_config.serialization_method, 'JSON')



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig


@dataclass
class Config(BaseConfig):
    """
    https://docs.metaplane.dev/reference/getting-started
    """

    api_token: str



import multiprocessing


def start_session_and_run(*target_args):
    from mage_ai.orchestration.db import db_connection

    if len(target_args) == 0:
        return None
    target = target_args[0]
    args = target_args[1:]

    db_connection.start_session(force=True)

    try:
        results = target(*args)
    finally:
        db_connection.close_session()
    return results


def create_process(target, args=()):
    from mage_ai.orchestration.db import engine

    new_args = [target, *args]
    engine.dispose()
    return multiprocessing.Process(target=start_session_and_run, args=new_args)



import traceback
from datetime import datetime, timedelta
from typing import Any, Dict, List, Tuple

import pytz
from dateutil.relativedelta import relativedelta

from mage_ai.data_integrations.utils.scheduler import (
    clear_source_output_files,
    initialize_state_and_runs,
)
from mage_ai.data_preparation.executors.executor_factory import ExecutorFactory
from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.data_preparation.logging.logger_manager_factory import LoggerManagerFactory
from mage_ai.data_preparation.models.block.utils import (
    create_block_runs_from_dynamic_block,
    dynamic_block_uuid,
    dynamic_block_values_and_metadata,
    is_dynamic_block,
    is_dynamic_block_child,
    should_reduce_output,
)
from mage_ai.data_preparation.models.constants import ExecutorType, PipelineType
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.pipelines.integration_pipeline import (
    IntegrationPipeline,
)
from mage_ai.data_preparation.models.triggers import (
    ScheduleInterval,
    ScheduleStatus,
    ScheduleType,
    get_triggers_by_pipeline,
)
from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.repo_manager import get_repo_config, get_repo_path
from mage_ai.data_preparation.sync import GitConfig
from mage_ai.data_preparation.sync.git_sync import GitSync
from mage_ai.data_preparation.variable_manager import get_global_variables
from mage_ai.orchestration.db import db_connection, safe_db_query
from mage_ai.orchestration.db.models.schedules import (
    Backfill,
    BlockRun,
    EventMatcher,
    PipelineRun,
    PipelineSchedule,
)
from mage_ai.orchestration.job_manager import JobType, job_manager
from mage_ai.orchestration.metrics.pipeline_run import calculate_metrics
from mage_ai.orchestration.notification.config import NotificationConfig
from mage_ai.orchestration.notification.sender import NotificationSender
from mage_ai.orchestration.utils.resources import get_compute, get_memory
from mage_ai.server.logger import Logger
from mage_ai.shared.array import find
from mage_ai.shared.constants import ENV_PROD
from mage_ai.shared.dates import compare
from mage_ai.shared.hash import index_by, merge_dict
from mage_ai.shared.retry import retry

MEMORY_USAGE_MAXIMUM = 0.95

logger = Logger().new_server_logger(__name__)


class PipelineScheduler:
    def __init__(
        self,
        pipeline_run: PipelineRun,
    ) -> None:
        self.pipeline_run = pipeline_run
        self.pipeline_schedule = pipeline_run.pipeline_schedule
        self.pipeline = Pipeline.get(pipeline_run.pipeline_uuid)
        self.logger_manager = LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline.uuid,
            partition=self.pipeline_run.execution_partition,
            repo_config=self.pipeline.repo_config,
        )
        self.logger = DictLogger(self.logger_manager.logger)
        self.notification_sender = NotificationSender(
            NotificationConfig.load(config=self.pipeline.repo_config.notification_config),
        )

        self.allow_blocks_to_fail = (
            self.pipeline_schedule.get_settings().allow_blocks_to_fail
            if self.pipeline_schedule else False
        )

    def start(self, should_schedule: bool = True) -> None:
        if get_preferences().sync_config:
            sync_config = GitConfig.load(config=get_preferences().sync_config)
            if sync_config.sync_on_pipeline_run:
                sync = GitSync(sync_config)
                sync.sync_data()

        if self.pipeline_run.status == PipelineRun.PipelineRunStatus.RUNNING:
            return
        self.pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        if should_schedule:
            self.schedule()

    def stop(self) -> None:
        stop_pipeline_run(
            self.pipeline_run,
            self.pipeline,
        )

    def schedule(self, block_runs: List[BlockRun] = None) -> None:
        self.__run_heartbeat()

        for b in self.pipeline_run.block_runs:
            b.refresh()

        if PipelineType.STREAMING == self.pipeline.type:
            self.__schedule_pipeline()
        else:
            if self.pipeline_run.all_blocks_completed(self.allow_blocks_to_fail):
                if PipelineType.INTEGRATION == self.pipeline.type:
                    tags = dict(
                        pipeline_run_id=self.pipeline_run.id,
                        pipeline_uuid=self.pipeline.uuid,
                    )
                    self.logger.info(
                        f'Calculate metrics for pipeline run {self.pipeline_run.id} started.',
                        **tags,
                    )
                    calculate_metrics(self.pipeline_run)
                    self.logger.info(
                        f'Calculate metrics for pipeline run {self.pipeline_run.id} completed.',
                        **merge_dict(tags, dict(metrics=self.pipeline_run.metrics)),
                    )

                if self.pipeline_run.any_blocks_failed():
                    self.pipeline_run.update(
                        status=PipelineRun.PipelineRunStatus.FAILED,
                        completed_at=datetime.now(),
                    )
                    self.notification_sender.send_pipeline_run_failure_message(
                        pipeline=self.pipeline,
                        pipeline_run=self.pipeline_run,
                    )
                else:
                    self.pipeline_run.complete()
                    self.notification_sender.send_pipeline_run_success_message(
                        pipeline=self.pipeline,
                        pipeline_run=self.pipeline_run,
                    )

                self.logger_manager.output_logs_to_destination()

                schedule = PipelineSchedule.get(
                    self.pipeline_run.pipeline_schedule_id,
                )

                if schedule:
                    backfills = schedule.backfills
                    # When all pipeline runs that are associated with backfill is done
                    if len(backfills) >= 1:
                        backfill = backfills[0]
                        if all([PipelineRun.PipelineRunStatus.COMPLETED == pr.status
                                for pr in backfill.pipeline_runs]):
                            backfill.update(
                                completed_at=datetime.now(),
                                status=Backfill.Status.COMPLETED,
                            )
                            schedule.update(
                                status=ScheduleStatus.INACTIVE,
                            )
                    # If running once, update the schedule to inactive when pipeline run is done
                    elif schedule.status == ScheduleStatus.ACTIVE and \
                            schedule.schedule_type == ScheduleType.TIME and \
                            schedule.schedule_interval == ScheduleInterval.ONCE:

                        schedule.update(status=ScheduleStatus.INACTIVE)
            elif PipelineType.INTEGRATION == self.pipeline.type:
                self.__schedule_integration_pipeline(block_runs)
            else:
                self.__schedule_blocks(block_runs)

    def on_block_complete(self, block_uuid: str) -> None:
        block_run = BlockRun.get(pipeline_run_id=self.pipeline_run.id, block_uuid=block_uuid)

        @retry(retries=3, delay=5)
        def update_status():
            block_run.update(
                status=BlockRun.BlockRunStatus.COMPLETED,
                completed_at=datetime.now(),
            )

        update_status()

        self.logger.info(
            f'BlockRun {block_run.id} (block_uuid: {block_uuid}) completes.',
            **self.__build_tags(
                block_run_id=block_run.id,
                block_uuid=block_run.block_uuid,
            ),
        )

        self.pipeline_run.refresh()
        if self.pipeline_run.status != PipelineRun.PipelineRunStatus.RUNNING:
            return
        else:
            self.schedule()

    def on_block_complete_without_schedule(self, block_uuid: str) -> None:
        block = self.pipeline.get_block(block_uuid)
        if block and is_dynamic_block(block):
            create_block_runs_from_dynamic_block(block, self.pipeline_run, block_uuid=block_uuid)

        block_run = BlockRun.get(pipeline_run_id=self.pipeline_run.id, block_uuid=block_uuid)

        @retry(retries=3, delay=5)
        def update_status():
            block_run.update(
                status=BlockRun.BlockRunStatus.COMPLETED,
                completed_at=datetime.now(),
            )

        update_status()

        self.logger.info(
            f'BlockRun {block_run.id} (block_uuid: {block_uuid}) completes.',
            **self.__build_tags(
                block_run_id=block_run.id,
                block_uuid=block_run.block_uuid,
            ),
        )

    def on_block_failure(self, block_uuid: str, **kwargs) -> None:
        block_run = BlockRun.get(pipeline_run_id=self.pipeline_run.id, block_uuid=block_uuid)
        metrics = block_run.metrics or {}

        @retry(retries=3, delay=5)
        def update_status():
            block_run.update(
                metrics=metrics,
                status=BlockRun.BlockRunStatus.FAILED,
            )
            if not self.allow_blocks_to_fail:
                self.pipeline_run.update(
                    status=PipelineRun.PipelineRunStatus.FAILED)

        update_status()

        if 'error' in kwargs:
            metrics['error'] = kwargs['error']

        tags = self.__build_tags(
            block_run_id=block_run.id,
            block_uuid=block_run.block_uuid,
        )

        self.logger.info(
            f'BlockRun {block_run.id} (block_uuid: {block_uuid}) failed.',
            **tags,
        )

        if not self.allow_blocks_to_fail:
            self.notification_sender.send_pipeline_run_failure_message(
                pipeline=self.pipeline,
                pipeline_run=self.pipeline_run,
            )

            if PipelineType.INTEGRATION == self.pipeline.type:
                # If a block/stream fails, stop all other streams
                job_manager.kill_pipeline_run_job(self.pipeline_run.id)

                self.logger.info(
                    f'Calculate metrics for pipeline run {self.pipeline_run.id} error started.',
                    tags=tags,
                )
                calculate_metrics(self.pipeline_run)
                self.logger.info(
                    f'Calculate metrics for pipeline run {self.pipeline_run.id} error completed.',
                    tags=merge_dict(tags, dict(metrics=self.pipeline_run.metrics)),
                )

    def memory_usage_failure(self, tags: Dict = {}) -> None:
        msg = 'Memory usage across all pipeline runs has reached or exceeded the maximum '\
            f'limit of {int(MEMORY_USAGE_MAXIMUM * 100)}%.'
        self.logger.info(msg, tags=tags)

        self.stop()

        self.notification_sender.send_pipeline_run_failure_message(
            pipeline=self.pipeline,
            pipeline_run=self.pipeline_run,
            message=msg,
        )

        if PipelineType.INTEGRATION == self.pipeline.type:
            self.logger.info(
                f'Calculate metrics for pipeline run {self.pipeline_run.id} error started.',
                tags=tags,
            )
            calculate_metrics(self.pipeline_run)
            self.logger.info(
                f'Calculate metrics for pipeline run {self.pipeline_run.id} error completed.',
                tags=merge_dict(tags, dict(metrics=self.pipeline_run.metrics)),
            )

    @property
    def initial_block_runs(self) -> List[BlockRun]:
        return [b for b in self.pipeline_run.block_runs
                if b.status == BlockRun.BlockRunStatus.INITIAL]

    @property
    def completed_block_runs(self) -> List[BlockRun]:
        return [b for b in self.pipeline_run.block_runs
                if b.status == BlockRun.BlockRunStatus.COMPLETED]

    @property
    def failed_block_runs(self) -> List[BlockRun]:
        return [b for b in self.pipeline_run.block_runs
                if b.status == BlockRun.BlockRunStatus.FAILED]

    @property
    def executable_block_runs(self) -> List[BlockRun]:
        completed_block_uuids = set(b.block_uuid for b in self.completed_block_runs)
        finished_block_uuids = set(
            b.block_uuid for b in self.pipeline_run.block_runs
            if b.status in [
                BlockRun.BlockRunStatus.COMPLETED,
                BlockRun.BlockRunStatus.UPSTREAM_FAILED,
                BlockRun.BlockRunStatus.FAILED,
            ]
        )

        executable_block_runs = list()
        for block_run in self.initial_block_runs:
            completed = False

            dynamic_upstream_block_uuids = block_run.metrics and block_run.metrics.get(
                'dynamic_upstream_block_uuids',
            )

            if dynamic_upstream_block_uuids:
                if self.allow_blocks_to_fail:
                    completed = all(uuid in finished_block_uuids
                                    for uuid in dynamic_upstream_block_uuids)
                else:
                    completed = all(uuid in completed_block_uuids
                                    for uuid in dynamic_upstream_block_uuids)
            else:
                block = self.pipeline.get_block(block_run.block_uuid)
                completed = block is not None and \
                    block.all_upstream_blocks_completed(completed_block_uuids)

            if completed:
                executable_block_runs.append(block_run)

        return executable_block_runs

    def __update_block_run_statuses(self, block_runs: List[BlockRun]) -> None:
        failed_block_uuids = set(
            b.block_uuid for b in self.pipeline_run.block_runs
            if b.status in [
                BlockRun.BlockRunStatus.UPSTREAM_FAILED,
                BlockRun.BlockRunStatus.FAILED,
            ]
        )
        not_updated_block_runs = []
        for block_run in block_runs:
            updated_status = False
            dynamic_upstream_block_uuids = block_run.metrics and block_run.metrics.get(
                'dynamic_upstream_block_uuids',
            )

            if dynamic_upstream_block_uuids:
                if all(
                    b in failed_block_uuids
                    for b in dynamic_upstream_block_uuids
                ):
                    block_run.update(
                        status=BlockRun.BlockRunStatus.UPSTREAM_FAILED)
                    updated_status = True
            else:
                block = self.pipeline.get_block(block_run.block_uuid)
                if any(
                    b in failed_block_uuids
                    for b in block.upstream_block_uuids
                ):
                    block_run.update(
                        status=BlockRun.BlockRunStatus.UPSTREAM_FAILED)
                    updated_status = True

            if not updated_status:
                not_updated_block_runs.append(block_run)

        self.pipeline_run.refresh()
        # keep iterating through block runs until no more updates can be made
        if len(block_runs) != len(not_updated_block_runs):
            self.__update_block_run_statuses(not_updated_block_runs)

    def __schedule_blocks(self, block_runs: List[BlockRun] = None) -> None:
        self.__update_block_run_statuses(self.initial_block_runs)
        block_runs_to_schedule = \
            self.executable_block_runs if block_runs is None else block_runs
        block_runs_to_schedule = \
            self.__fetch_crashed_block_runs() + block_runs_to_schedule

        for b in block_runs_to_schedule:
            tags = dict(
                block_run_id=b.id,
                block_uuid=b.block_uuid,
            )

            b.update(
                started_at=datetime.now(),
                status=BlockRun.BlockRunStatus.QUEUED,
            )

            job_manager.add_job(
                JobType.BLOCK_RUN,
                b.id,
                run_block,
                # args
                self.pipeline_run.id,
                b.id,
                get_variables(self.pipeline_run),
                self.__build_tags(**tags),
            )

    def __schedule_integration_pipeline(self, block_runs: List[BlockRun] = None) -> None:
        if job_manager.has_pipeline_run_job(self.pipeline_run.id):
            return

        block_runs_to_schedule = self.initial_block_runs if block_runs is None else block_runs
        block_runs_to_schedule = self.__fetch_crashed_block_runs() + block_runs_to_schedule

        if len(block_runs_to_schedule) > 0:
            self.logger.info(
                f'Start a process for PipelineRun {self.pipeline_run.id}',
                **self.__build_tags(),
            )

            job_manager.add_job(
                JobType.PIPELINE_RUN,
                self.pipeline_run.id,
                run_integration_pipeline,
                # args
                self.pipeline_run.id,
                [b.id for b in block_runs_to_schedule],
                get_variables(self.pipeline_run, dict(
                    pipeline_uuid=self.pipeline.uuid,
                )),
                self.__build_tags(),
            )

    def __schedule_pipeline(self) -> None:
        if job_manager.has_pipeline_run_job(self.pipeline_run.id):
            return
        self.logger.info(
            f'Start a process for PipelineRun {self.pipeline_run.id}',
            **self.__build_tags(),
        )
        job_manager.add_job(
            JobType.PIPELINE_RUN,
            self.pipeline_run.id,
            run_pipeline,
            # args
            self.pipeline_run.id,
            get_variables(self.pipeline_run),
            self.__build_tags(),
        )

    def __fetch_crashed_block_runs(self) -> None:
        running_or_queued_block_runs = [b for b in self.pipeline_run.block_runs if b.status in [
            BlockRun.BlockRunStatus.RUNNING,
            BlockRun.BlockRunStatus.QUEUED,
        ]]

        crashed_runs = []
        for br in running_or_queued_block_runs:
            if not job_manager.has_block_run_job(br.id):
                br.update(status=BlockRun.BlockRunStatus.INITIAL)
                crashed_runs.append(br)

        return crashed_runs

    def __build_tags(self, **kwargs):
        return merge_dict(kwargs, dict(
            pipeline_run_id=self.pipeline_run.id,
            pipeline_schedule_id=self.pipeline_run.pipeline_schedule_id,
            pipeline_uuid=self.pipeline.uuid,
        ))

    def __run_heartbeat(self) -> None:
        load1, load5, load15, cpu_count = get_compute()
        cpu_usage = load15 / cpu_count if cpu_count else None

        free_memory, used_memory, total_memory = get_memory()
        memory_usage = used_memory / total_memory if total_memory else None

        tags = dict(
            cpu=load15,
            cpu_total=cpu_count,
            cpu_usage=cpu_usage,
            memory=used_memory,
            memory_total=total_memory,
            memory_usage=memory_usage,
            pipeline_run_id=self.pipeline_run.id,
            pipeline_uuid=self.pipeline.uuid,
        )

        self.logger.info(
            f'Pipeline {self.pipeline.uuid} for run {self.pipeline_run.id} '
            f'in schedule {self.pipeline_run.pipeline_schedule_id} is alive.',
            **tags,
        )

        if memory_usage and memory_usage >= MEMORY_USAGE_MAXIMUM:
            self.memory_usage_failure(tags)


def run_integration_pipeline(
    pipeline_run_id: int,
    executable_block_runs: List[int],
    variables: Dict,
    tags: Dict,
) -> None:
    pipeline_run = PipelineRun.query.get(pipeline_run_id)
    pipeline_scheduler = PipelineScheduler(pipeline_run)
    integration_pipeline = IntegrationPipeline.get(pipeline_scheduler.pipeline.uuid)
    pipeline_scheduler.logger.info(f'Execute PipelineRun {pipeline_run.id}: '
                                   f'pipeline {integration_pipeline.uuid}',
                                   **tags)

    pipeline_scheduler.logger.info(f'Executable block runs: {executable_block_runs}',
                                   **tags)

    all_block_runs = BlockRun.query.filter(BlockRun.pipeline_run_id == pipeline_run_id)
    block_runs = list(filter(lambda br: br.id in executable_block_runs, all_block_runs))

    pipeline_schedule = pipeline_run.pipeline_schedule
    schedule_interval = pipeline_schedule.schedule_interval
    if ScheduleType.API == pipeline_schedule.schedule_type:
        execution_date = datetime.utcnow()
    else:
        # This will be none if trigger is API type
        execution_date = pipeline_schedule.current_execution_date()

    end_date = None
    start_date = None
    date_diff = None

    if ScheduleInterval.ONCE == schedule_interval:
        end_date = variables.get('_end_date')
        start_date = variables.get('_start_date')
    elif ScheduleInterval.HOURLY == schedule_interval:
        date_diff = timedelta(hours=1)
    elif ScheduleInterval.DAILY == schedule_interval:
        date_diff = timedelta(days=1)
    elif ScheduleInterval.WEEKLY == schedule_interval:
        date_diff = timedelta(weeks=1)
    elif ScheduleInterval.MONTHLY == schedule_interval:
        date_diff = relativedelta(months=1)

    if date_diff is not None:
        end_date = (execution_date).isoformat()
        start_date = (execution_date - date_diff).isoformat()

    runtime_arguments = dict(
        _end_date=end_date,
        _execution_date=execution_date.isoformat(),
        _execution_partition=pipeline_run.execution_partition,
        _start_date=start_date,
    )

    data_loader_block = integration_pipeline.data_loader
    data_exporter_block = integration_pipeline.data_exporter
    executable_block_run_ids = set(executable_block_runs)

    for stream in integration_pipeline.streams(variables):
        tap_stream_id = stream['tap_stream_id']
        destination_table = stream.get('destination_table', tap_stream_id)

        block_runs_for_stream = list(filter(lambda br: tap_stream_id in br.block_uuid, block_runs))
        if len(block_runs_for_stream) == 0:
            continue

        indexes = [0]
        for br in block_runs_for_stream:
            parts = br.block_uuid.split(':')
            if len(parts) >= 3:
                indexes.append(int(parts[2]))
        max_index = max(indexes)

        all_block_runs_for_stream = list(filter(
            lambda br: tap_stream_id in br.block_uuid,
            all_block_runs,
        ))
        all_indexes = [0]
        for br in all_block_runs_for_stream:
            parts = br.block_uuid.split(':')
            if len(parts) >= 3:
                all_indexes.append(int(parts[2]))
        max_index_for_stream = max(all_indexes)

        for idx in range(max_index + 1):
            block_runs_in_order = []
            current_block = data_loader_block

            while True:
                block_runs_in_order.append(
                    find(
                        lambda b: b.block_uuid == f'{current_block.uuid}:{tap_stream_id}:{idx}',
                        all_block_runs,
                    )
                )
                downstream_blocks = current_block.downstream_blocks
                if len(downstream_blocks) == 0:
                    break
                current_block = downstream_blocks[0]

            data_loader_uuid = f'{data_loader_block.uuid}:{tap_stream_id}:{idx}'
            data_exporter_uuid = f'{data_exporter_block.uuid}:{tap_stream_id}:{idx}'

            data_loader_block_run = find(
                lambda b: b.block_uuid == data_loader_uuid,
                all_block_runs,
            )
            data_exporter_block_run = find(
                lambda b: b.block_uuid == data_exporter_uuid,
                block_runs_for_stream,
            )
            if not data_loader_block_run or not data_exporter_block_run:
                continue

            transformer_block_runs = [br for br in block_runs_in_order if (
                br.block_uuid not in [data_loader_uuid, data_exporter_uuid] and
                br.id in executable_block_run_ids
            )]

            index = stream.get('index', idx)

            shared_dict = dict(
                destination_table=destination_table,
                index=index,
                is_last_block_run=(index == max_index_for_stream),
                selected_streams=[
                    tap_stream_id,
                ],
            )
            block_runs_and_configs = [
                (data_loader_block_run, shared_dict),
            ] + [(br, shared_dict) for br in transformer_block_runs] + [
                (data_exporter_block_run, shared_dict),
            ]
            if len(executable_block_runs) == 1 and \
                    data_exporter_block_run.id in executable_block_run_ids:
                block_runs_and_configs = block_runs_and_configs[-1:]
            elif data_loader_block_run.id not in executable_block_run_ids:
                block_runs_and_configs = block_runs_and_configs[1:]

            block_failed = False
            for idx2, tup in enumerate(block_runs_and_configs):
                block_run, template_runtime_configuration = tup

                tags_updated = merge_dict(tags, dict(
                    block_run_id=block_run.id,
                    block_uuid=block_run.block_uuid,
                ))

                if block_failed:
                    block_run.update(
                        status=BlockRun.BlockRunStatus.UPSTREAM_FAILED,
                    )
                    continue

                pipeline_run.refresh()
                if pipeline_run.status != PipelineRun.PipelineRunStatus.RUNNING:
                    return

                block_run.update(
                    started_at=datetime.now(),
                    status=BlockRun.BlockRunStatus.RUNNING,
                )
                pipeline_scheduler.logger.info(
                    f'Start a process for BlockRun {block_run.id}',
                    **tags_updated,
                )

                try:
                    run_block(
                        pipeline_run_id,
                        block_run.id,
                        variables,
                        tags_updated,
                        pipeline_type=PipelineType.INTEGRATION,
                        verify_output=False,
                        retry_count=1,  # Not retry for data integration pipeline blocks
                        runtime_arguments=runtime_arguments,
                        schedule_after_complete=False,
                        template_runtime_configuration=template_runtime_configuration,
                    )
                except Exception as e:
                    if pipeline_scheduler.allow_blocks_to_fail:
                        block_failed = True
                    else:
                        raise e
                else:
                    if f'{data_loader_block.uuid}:{tap_stream_id}' in block_run.block_uuid or \
                            f'{data_exporter_block.uuid}:{tap_stream_id}' in block_run.block_uuid:

                        tags2 = merge_dict(tags_updated.get('tags', {}), dict(
                            destination_table=destination_table,
                            index=index,
                            stream=tap_stream_id,
                        ))
                        pipeline_scheduler.logger.info(
                            f'Calculate metrics for pipeline run {pipeline_run.id} started.',
                            **tags_updated,
                            tags=tags2,
                        )
                        try:
                            calculate_metrics(pipeline_run)
                            pipeline_scheduler.logger.info(
                                f'Calculate metrics for pipeline run {pipeline_run.id} completed.',
                                **tags_updated,
                                tags=merge_dict(tags2, dict(metrics=pipeline_run.metrics)),
                            )
                        except Exception:
                            pipeline_scheduler.logger.error(
                                f'Failed to calculate metrics for pipeline run {pipeline_run.id}. '
                                f'{traceback.format_exc()}',
                                **tags_updated,
                                tags=tags2,
                            )


def run_block(
    pipeline_run_id,
    block_run_id,
    variables,
    tags,
    input_from_output: Dict = None,
    pipeline_type: PipelineType = None,
    verify_output: bool = True,
    retry_count: int = 3,
    runtime_arguments: Dict = None,
    schedule_after_complete: bool = False,
    template_runtime_configuration: Dict = None,
) -> Any:
    pipeline_run = PipelineRun.query.get(pipeline_run_id)
    if pipeline_run.status != PipelineRun.PipelineRunStatus.RUNNING:
        return {}

    pipeline_scheduler = PipelineScheduler(pipeline_run)

    pipeline = pipeline_scheduler.pipeline

    block_run = BlockRun.query.get(block_run_id)
    block_run.update(
        started_at=datetime.now(),
        status=BlockRun.BlockRunStatus.RUNNING,
    )

    pipeline_scheduler.logger.info(
        f'Execute PipelineRun {pipeline_run.id}, BlockRun {block_run.id}: '
        f'pipeline {pipeline.uuid} block {block_run.block_uuid}',
        **tags)

    if schedule_after_complete:
        on_complete = pipeline_scheduler.on_block_complete
    else:
        on_complete = pipeline_scheduler.on_block_complete_without_schedule

    block_run_data = block_run.metrics or {}
    dynamic_block_index = block_run_data.get('dynamic_block_index', None)
    dynamic_upstream_block_uuids = block_run_data.get('dynamic_upstream_block_uuids', None)

    execution_partition = pipeline_run.execution_partition

    block_uuid = block_run.block_uuid

    # If there are upstream blocks that were dynamically created, and if any of them are configured
    # to reduce their output, we must update the dynamic_upstream_block_uuids to include all
    # the upstream block’s dynamically created blocks by getting the upstream block’s
    # dynamic parent block and collecting that parent’s output.
    if dynamic_upstream_block_uuids:
        dynamic_upstream_block_uuids_reduce = []
        dynamic_upstream_block_uuids_no_reduce = []

        for upstream_block_uuid in dynamic_upstream_block_uuids:
            upstream_block = pipeline.get_block(upstream_block_uuid)

            if not should_reduce_output(upstream_block):
                dynamic_upstream_block_uuids_no_reduce.append(upstream_block_uuid)
                continue

            parts = upstream_block_uuid.split(':')
            suffix = None
            if len(parts) >= 3:
                # A block can have a UUID such as: some_uuid:0 or some_uuid:0:1
                suffix = ':'.join(parts[2:])

            # We currently limit a block to only have 1 direct dynamic parent.
            # We are looping over the upstream blocks just in case we support having
            # multiple direct dynamic parents.
            for block_grandparent in list(filter(
                lambda x: is_dynamic_block(x),
                upstream_block.upstream_blocks,
            )):

                block_grandparent_uuid = block_grandparent.uuid

                if suffix and is_dynamic_block_child(block_grandparent):
                    block_grandparent_uuid = f'{block_grandparent_uuid}:{suffix}'

                values, block_metadata = dynamic_block_values_and_metadata(
                    block_grandparent,
                    execution_partition,
                    block_grandparent_uuid,
                )

                for idx, value in enumerate(values):
                    if idx < len(block_metadata):
                        metadata = block_metadata[idx].copy()
                    else:
                        metadata = {}

                    dynamic_upstream_block_uuids_reduce.append(
                        dynamic_block_uuid(
                            upstream_block.uuid,
                            metadata,
                            idx,
                            upstream_block_uuid=block_grandparent_uuid,
                        ))

        dynamic_upstream_block_uuids = dynamic_upstream_block_uuids_reduce + \
            dynamic_upstream_block_uuids_no_reduce

    return ExecutorFactory.get_block_executor(
        pipeline,
        block_uuid,
        execution_partition=execution_partition,
    ).execute(
        analyze_outputs=False,
        block_run_id=block_run.id,
        global_vars=variables,
        update_status=False,
        on_complete=on_complete,
        on_failure=pipeline_scheduler.on_block_failure,
        tags=tags,
        input_from_output=input_from_output,
        verify_output=verify_output,
        pipeline_run_id=pipeline_run_id,
        retry_count=retry_count,
        runtime_arguments=runtime_arguments,
        template_runtime_configuration=template_runtime_configuration,
        dynamic_block_index=dynamic_block_index,
        dynamic_block_uuid=None if dynamic_block_index is None else block_run.block_uuid,
        dynamic_upstream_block_uuids=dynamic_upstream_block_uuids,
    )


def run_pipeline(pipeline_run_id, variables, tags):
    pipeline_run = PipelineRun.query.get(pipeline_run_id)
    pipeline_scheduler = PipelineScheduler(pipeline_run)
    pipeline = pipeline_scheduler.pipeline
    pipeline_scheduler.logger.info(f'Execute PipelineRun {pipeline_run.id}: '
                                   f'pipeline {pipeline.uuid}',
                                   **tags)
    executor_type = ExecutorFactory.get_pipeline_executor_type(pipeline)
    try:
        pipeline_run.update(executor_type=executor_type)
    except Exception:
        traceback.print_exc()
    ExecutorFactory.get_pipeline_executor(
        pipeline,
        execution_partition=pipeline_run.execution_partition,
        executor_type=executor_type,
    ).execute(
        global_vars=variables,
        pipeline_run_id=pipeline_run_id,
        tags=tags,
    )


def configure_pipeline_run_payload(
    pipeline_schedule: PipelineSchedule,
    pipeline_type: PipelineType,
    payload: Dict = {},
) -> Tuple[Dict, bool]:
    if 'variables' not in payload:
        payload['variables'] = {}

    payload['pipeline_schedule_id'] = pipeline_schedule.id
    payload['pipeline_uuid'] = pipeline_schedule.pipeline_uuid
    execution_date = payload.get('execution_date')
    if execution_date is None:
        payload['execution_date'] = datetime.utcnow()
    elif not isinstance(execution_date, datetime):
        payload['execution_date'] = datetime.fromisoformat(execution_date)

    is_integration = PipelineType.INTEGRATION == pipeline_type
    if is_integration:
        payload['create_block_runs'] = False

    return payload, is_integration


def start_scheduler(pipeline_run: PipelineRun) -> None:
    from mage_ai.orchestration.pipeline_scheduler import (
        PipelineScheduler,
        get_variables,
    )

    pipeline_scheduler = PipelineScheduler(pipeline_run)
    is_integration = PipelineType.INTEGRATION == pipeline_run.pipeline.type

    if is_integration:
        initialize_state_and_runs(
            pipeline_run,
            pipeline_scheduler.logger,
            get_variables(pipeline_run),
        )
    else:
        pipeline_run.create_block_runs()

    pipeline_scheduler.start(should_schedule=False)


@safe_db_query
def retry_pipeline_run(
    pipeline_run: Dict,
) -> 'PipelineRun':
    pipeline_uuid = pipeline_run['pipeline_uuid']
    pipeline = Pipeline.get(pipeline_uuid, check_if_exists=True)
    if pipeline is None or not pipeline.is_valid_pipeline(pipeline.dir_path):
        raise Exception(f'Pipeline {pipeline_uuid} is not a valid pipeline.')

    pipeline_schedule_id = pipeline_run['pipeline_schedule_id']
    pipeline_run_model = PipelineRun(
        id=pipeline_run['id'],
        pipeline_schedule_id=pipeline_schedule_id,
        pipeline_uuid=pipeline_uuid,
    )
    execution_date = datetime.fromisoformat(pipeline_run['execution_date'])
    new_pipeline_run = pipeline_run_model.create(
        backfill_id=pipeline_run.get('backfill_id'),
        create_block_runs=False,
        execution_date=execution_date,
        event_variables=pipeline_run.get('event_variables', {}),
        pipeline_schedule_id=pipeline_schedule_id,
        pipeline_uuid=pipeline_run_model.pipeline_uuid,
        variables=pipeline_run.get('variables', {}),
    )
    start_scheduler(new_pipeline_run)

    return new_pipeline_run


def stop_pipeline_run(
    pipeline_run: PipelineRun,
    pipeline: Pipeline = None,
) -> None:
    if pipeline_run.status not in [PipelineRun.PipelineRunStatus.INITIAL,
                                   PipelineRun.PipelineRunStatus.RUNNING]:
        return

    pipeline_run.update(status=PipelineRun.PipelineRunStatus.CANCELLED)

    # Cancel all the block runs
    block_runs_to_cancel = []
    running_blocks = []
    for b in pipeline_run.block_runs:
        if b.status in [
            BlockRun.BlockRunStatus.INITIAL,
            BlockRun.BlockRunStatus.QUEUED,
            BlockRun.BlockRunStatus.RUNNING,
        ]:
            block_runs_to_cancel.append(b)
        if b.status == BlockRun.BlockRunStatus.RUNNING:
            running_blocks.append(b)
    BlockRun.batch_update_status(
        [b.id for b in block_runs_to_cancel],
        BlockRun.BlockRunStatus.CANCELLED,
    )

    if pipeline and pipeline.type in [PipelineType.INTEGRATION, PipelineType.STREAMING]:
        job_manager.kill_pipeline_run_job(pipeline_run.id)
        if pipeline_run.executor_type == ExecutorType.K8S:
            """
            TODO: Support running and cancelling pipeline runs in ECS and GCP_CLOUD_RUN executors
            """
            ExecutorFactory.get_pipeline_executor(
                pipeline,
                executor_type=pipeline_run.executor_type,
            ).cancel(pipeline_run_id=pipeline_run.id)
    else:
        for b in running_blocks:
            job_manager.kill_block_run_job(b.id)


def check_sla():
    repo_pipelines = set(Pipeline.get_all_pipelines(get_repo_path()))
    pipeline_schedules = \
        set([
            s.id
            for s in PipelineSchedule.active_schedules(pipeline_uuids=repo_pipelines)
        ])

    pipeline_runs = PipelineRun.in_progress_runs(pipeline_schedules)

    if pipeline_runs:
        notification_sender = NotificationSender(
             NotificationConfig.load(config=get_repo_config(get_repo_path()).notification_config),
        )

        current_time = datetime.now(tz=pytz.UTC)
        # TODO: combine all SLA alerts in one notification
        for pipeline_run in pipeline_runs:
            sla = pipeline_run.pipeline_schedule.sla
            if not sla:
                continue
            start_date = \
                pipeline_run.execution_date \
                if pipeline_run.execution_date is not None \
                else pipeline_run.created_at
            if compare(start_date, current_time - timedelta(seconds=sla)) == 1:
                # passed SLA for pipeline_run
                notification_sender.send_pipeline_run_sla_passed_message(
                    Pipeline.get(pipeline_run.pipeline_schedule.pipeline_uuid),
                    pipeline_run,
                )

                pipeline_run.update(passed_sla=True)


def schedule_all():
    """
    1. Check whether any new pipeline runs need to be scheduled.
    2. In active pipeline runs, check whether any block runs need to be scheduled.
    """
    db_connection.session.expire_all()

    repo_pipelines = set(Pipeline.get_all_pipelines(get_repo_path()))

    sync_schedules(list(repo_pipelines))

    active_pipeline_schedules = \
        list(PipelineSchedule.active_schedules(pipeline_uuids=repo_pipelines))

    backfills = Backfill.filter(pipeline_schedule_ids=[ps.id for ps in active_pipeline_schedules])

    backfills_by_pipeline_schedule_id = index_by(
        lambda backfill: backfill.pipeline_schedule_id,
        backfills,
    )

    for pipeline_schedule in active_pipeline_schedules:
        if pipeline_schedule.should_schedule() and \
                pipeline_schedule.id not in backfills_by_pipeline_schedule_id:

            pipeline_uuid = pipeline_schedule.pipeline_uuid
            payload = dict(
                execution_date=pipeline_schedule.current_execution_date(),
                pipeline_schedule_id=pipeline_schedule.id,
                pipeline_uuid=pipeline_uuid,
                variables=pipeline_schedule.variables,
            )

            pipeline = Pipeline.get(pipeline_uuid)
            is_integration = PipelineType.INTEGRATION == pipeline.type
            if is_integration:
                payload['create_block_runs'] = False

            running_pipeline_run = find(
                lambda r: r.status in [
                    PipelineRun.PipelineRunStatus.INITIAL,
                    PipelineRun.PipelineRunStatus.RUNNING,
                ],
                pipeline_schedule.pipeline_runs
            )
            if pipeline_schedule.get_settings().skip_if_previous_running and \
                    running_pipeline_run is not None:

                payload['create_block_runs'] = False
                pipeline_run = PipelineRun.create(**payload)
                pipeline_run.update(status=PipelineRun.PipelineRunStatus.CANCELLED)
            else:
                pipeline_run = PipelineRun.create(**payload)
                pipeline_scheduler = PipelineScheduler(pipeline_run)
                if is_integration:
                    block_runs = BlockRun.query.filter(
                        BlockRun.pipeline_run_id == pipeline_run.id).all()
                    if len(block_runs) == 0:
                        clear_source_output_files(
                            pipeline_run,
                            pipeline_scheduler.logger,
                        )
                        initialize_state_and_runs(
                            pipeline_run,
                            pipeline_scheduler.logger,
                            get_variables(pipeline_run),
                        )

                pipeline_scheduler.start(should_schedule=False)

    active_pipeline_runs = PipelineRun.active_runs(
        pipeline_uuids=repo_pipelines,
        include_block_runs=True,
    )
    logger.info(f'Active pipeline runs: {[p.id for p in active_pipeline_runs]}')

    for r in active_pipeline_runs:
        try:
            r.refresh()
            PipelineScheduler(r).schedule()
        except Exception:
            logger.exception(f'Failed to schedule {r}')
            traceback.print_exc()
            continue
    job_manager.clean_up_jobs()


def schedule_with_event(event: Dict = dict()):
    logger.info(f'Schedule with event {event}')
    all_event_matchers = EventMatcher.active_event_matchers()
    for e in all_event_matchers:
        if e.match(event):
            logger.info(f'Event matched with {e}')
            pipeline_schedules = e.active_pipeline_schedules()
            for p in pipeline_schedules:
                payload = dict(
                    execution_date=datetime.now(),
                    pipeline_schedule_id=p.id,
                    pipeline_uuid=p.pipeline_uuid,
                    variables=merge_dict(p.variables or dict(), dict(event=event)),
                )
                pipeline_run = PipelineRun.create(**payload)
                PipelineScheduler(pipeline_run).start(should_schedule=True)
        else:
            logger.info(f'Event not matched with {e}')


def sync_schedules(pipeline_uuids: List[str]):
    # Sync schedule configs from triggers.yaml to DB
    for pipeline_uuid in pipeline_uuids:
        pipeline_triggers = get_triggers_by_pipeline(pipeline_uuid)

        logger.debug(f'Sync pipeline trigger configs for {pipeline_uuid}: {pipeline_triggers}.')
        for pipeline_trigger in pipeline_triggers:
            PipelineSchedule.create_or_update(pipeline_trigger)


def get_variables(pipeline_run, extra_variables: Dict = {}) -> Dict:
    if not pipeline_run:
        return {}

    pipeline_run_variables = pipeline_run.variables or {}
    event_variables = pipeline_run.event_variables or {}

    variables = merge_dict(
        merge_dict(
            get_global_variables(pipeline_run.pipeline_uuid) or dict(),
            pipeline_run.pipeline_schedule.variables or dict(),
        ),
        pipeline_run_variables,
    )

    # For backwards compatibility
    for k, v in event_variables.items():
        if k not in variables:
            variables[k] = v

    if pipeline_run.execution_date:
        variables['ds'] = pipeline_run.execution_date.strftime('%Y-%m-%d')
        variables['hr'] = pipeline_run.execution_date.strftime('%H')

    variables['env'] = ENV_PROD
    variables['event'] = merge_dict(variables.get('event', {}), event_variables)
    variables['execution_date'] = pipeline_run.execution_date
    variables['execution_partition'] = pipeline_run.execution_partition
    variables.update(extra_variables)

    return variables



from dataclasses import dataclass, field
from enum import Enum
from mage_ai.services.email.config import EmailConfig
from mage_ai.services.google_chat.config import GoogleChatConfig
from mage_ai.services.slack.config import SlackConfig
from mage_ai.services.teams.config import TeamsConfig
from mage_ai.services.opsgenie.config import OpsgenieConfig
from mage_ai.shared.config import BaseConfig
from typing import Dict, List
import traceback


class AlertOn(str, Enum):
    PIPELINE_RUN_FAILURE = 'trigger_failure'
    PIPELINE_RUN_SUCCESS = 'trigger_success'
    PIPELINE_RUN_PASSED_SLA = 'trigger_passed_sla'


DEFAULT_ALERT_ON = [
    AlertOn.PIPELINE_RUN_FAILURE,
    AlertOn.PIPELINE_RUN_PASSED_SLA,
]


@dataclass
class NotificationConfig(BaseConfig):
    alert_on: List[AlertOn] = field(default_factory=lambda: DEFAULT_ALERT_ON)
    email_config: EmailConfig = None
    google_chat_config: GoogleChatConfig = None
    opsgenie_config: OpsgenieConfig = None
    slack_config: SlackConfig = None
    teams_config: TeamsConfig = None

    @classmethod
    def load(self, config_path: str = None, config: Dict = None):
        notification_config = super().load(config_path=config_path, config=config)
        if notification_config.slack_config is not None and \
                type(notification_config.slack_config) is dict:
            try:
                notification_config.slack_config = SlackConfig.load(
                    config=notification_config.slack_config,
                )
            except Exception:
                traceback.print_exc()
                notification_config.slack_config = None
        if notification_config.teams_config is not None and \
                type(notification_config.teams_config) is dict:
            try:
                notification_config.teams_config = TeamsConfig.load(
                    config=notification_config.teams_config,
                )
            except Exception:
                traceback.print_exc()
                notification_config.teams_config = None
        if notification_config.google_chat_config is not None and \
                type(notification_config.google_chat_config) is dict:
            try:
                notification_config.google_chat_config = GoogleChatConfig.load(
                    config=notification_config.google_chat_config,
                )
            except Exception:
                traceback.print_exc()
                notification_config.google_chat_config = None
        if notification_config.email_config is not None and \
                type(notification_config.email_config) is dict:
            try:
                notification_config.email_config = EmailConfig.load(
                    config=notification_config.email_config,
                )
            except Exception:
                traceback.print_exc()
                notification_config.email_config = None
        if notification_config.opsgenie_config is not None and \
                type(notification_config.opsgenie_config) is dict:
            try:
                notification_config.opsgenie_config = OpsgenieConfig.load(
                    config=notification_config.opsgenie_config,
                )
            except Exception:
                traceback.print_exc()
                notification_config.opsgenie_config = None
        return notification_config



from datetime import datetime
from mage_ai.server.logger import Logger

LOGGER = Logger().new_server_logger(__name__)


def debug(text):
    now = datetime.utcnow().isoformat()
    LOGGER.debug(f'[{now}][api.views] {text}')


def error(text):
    now = datetime.utcnow().isoformat()

    LOGGER.error(f'[{now}][api.views] {text}')


def info(text):
    now = datetime.utcnow().isoformat()
    LOGGER.info(f'[{now}][api.views] {text}')



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.git import Git
from mage_ai.data_preparation.preferences import get_preferences


class GitBranchResource(GenericResource):
    @classmethod
    def collection(self, query, meta, user, **kwargs):
        git_manager = Git.get_manager(user=user)
        return self.build_result_set(
            [dict(name=branch) for branch in git_manager.branches],
            user,
            **kwargs,
        )

    @classmethod
    def create(self, payload, user, **kwargs):
        branch = payload.get('name')
        git_manager = Git.get_manager(user=user)
        git_manager.switch_branch(branch)

        return self(dict(name=git_manager.current_branch), user, **kwargs)

    @classmethod
    async def member(self, pk, user, **kwargs):
        branch = None
        preferences = get_preferences(user=user)
        if preferences.has_valid_git_config():
            git_manager = Git.get_manager(user=user)
            branch = git_manager.current_branch
        return self(dict(name=branch), user, **kwargs)

    async def update(self, payload, **kwargs):
        git_manager = Git.get_manager(user=self.current_user)
        action_type = payload.get('action_type')
        if action_type == 'status':
            status = git_manager.status()
            untracked_files = git_manager.untracked_files()
            modified_files = git_manager.modified_files
            self.model = dict(
                name=git_manager.current_branch,
                status=status,
                untracked_files=untracked_files,
                modified_files=modified_files,
            )
        elif action_type == 'commit':
            message = payload.get('message')
            files = payload.get('files', None)
            if not message:
                error = ApiError.RESOURCE_ERROR
                error.update({
                    'message': 'Message is empty, please add a message for your commit.',
                })
                raise ApiError(error)
            git_manager.commit(message, files)
        elif action_type == 'push':
            git_manager.push()
        elif action_type == 'pull':
            git_manager.pull()
        elif action_type == 'reset':
            git_manager.reset()
        elif action_type == 'clone':
            git_manager.clone()

        return self



import os
import pandas as pd

DIR_PATH = os.path.abspath(os.path.dirname(__file__))


def list_dataset_names():
    paths = os.listdir(DIR_PATH)
    dataset_names = []
    for path in paths:
        if path.endswith('.csv'):
            dataset_names.append(path.split('/')[-1])
    return dataset_names


def load_dataset(name):
    fpath = os.path.join(DIR_PATH, name)
    if not os.path.isfile(fpath):
        raise Exception(f'Dataset {name} is not available')
    return pd.read_csv(fpath)


def load_titanic():
    return load_dataset('titanic_survival.csv')


def load_product_purchases():
    return load_dataset('product_purchases.csv')


def load_salary_survey():
    return load_dataset('salary_survey.csv')


def load_user_emails():
    return load_dataset('user_emails.csv')



import bcrypt


def generate_salt() -> str:
    return bcrypt.gensalt(14)


def create_bcrypt_hash(password: str, salt: str) -> str:
    password_bytes = password.encode()
    password_hash_bytes = bcrypt.hashpw(password_bytes, salt)
    password_hash_str = password_hash_bytes.decode()
    return password_hash_str


def verify_password(password: str, hash_from_database: str) -> bool:
    password_bytes = password.encode()
    hash_bytes = hash_from_database.encode()

    # this will automatically retrieve the salt from the hash,
    # then combine it with the password (parameter 1)
    # and then hash that, and compare it to the user's hash
    does_match = bcrypt.checkpw(password_bytes, hash_bytes)

    return does_match



from mage_ai import settings
from mage_ai.api.errors import ApiError
from mage_ai.api.operations.constants import META_KEY_LIMIT, META_KEY_OFFSET
from mage_ai.api.resources.BaseResource import BaseResource
from mage_ai.orchestration.db import db_connection, safe_db_query
from mage_ai.orchestration.db.errors import DoesNotExistError, ValidationError
from mage_ai.shared.hash import ignore_keys, merge_dict
from sqlalchemy.orm.query import Query
import inspect


class DatabaseResource(BaseResource):
    DEFAULT_LIMIT = 40

    @classmethod
    @safe_db_query
    async def process_collection(self, query, meta, user, **kwargs):
        limit = int(meta.get(META_KEY_LIMIT, self.DEFAULT_LIMIT))
        offset = int(meta.get(META_KEY_OFFSET, 0))

        total_results = self.collection(query, meta, user, **kwargs)
        if total_results and inspect.isawaitable(total_results):
            total_results = await total_results

        if issubclass(total_results.__class__, Query):
            total_count = total_results.count()

            results = total_results.limit(limit + 1).offset(offset).all()
        else:
            total_count = len(total_results)

            start_idx = offset
            end_idx = start_idx + limit

            results = total_results[start_idx:(end_idx + 1)]

        results_size = len(results)
        has_next = results_size > limit
        final_end_idx = results_size - 1 if has_next else results_size

        result_set = self.build_result_set(
            results[0:final_end_idx],
            user,
            **kwargs,
        )
        result_set.metadata = {
            'count': total_count,
            'next': has_next,
        }
        return result_set

    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        query = ignore_keys(query_arg, [settings.QUERY_API_KEY])
        parent_model = kwargs.get('parent_model')
        if parent_model and self.parent_resource():
            column_name, parent_class = next(
                (k, v) for k, v in self.parent_resource().items() if isinstance(
                    parent_model, v.model_class))
            where = {}
            where[column_name] = parent_model.id

            filters = []
            for col, val in merge_dict(query, where).items():
                filters.append(self.model_class)
            return self.model_class.query.filter(
                **merge_dict(query, where)).all()
        else:
            return self.model_class.query.filter(**query).all()

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        parent_model = kwargs.get('parent_model')
        column_name = None
        parent_class = None
        if parent_model and self.parent_models():
            try:
                column_name, parent_class = next(
                    (k, v) for k, v in self.parent_models().items() if isinstance(
                        parent_model, v))
                payload[column_name] = parent_model.id
            except StopIteration:
                pass
        if parent_model and self.parent_resource():
            try:
                column_name, parent_class = next(
                    (k, v) for k, v in self.parent_resource().items() if isinstance(
                        parent_model, v.model_class))
                payload[column_name] = parent_model.id
            except StopIteration:
                pass
        try:
            model = self.model_class(**payload)
            model.full_clean()
            model.save()
            self.create_associated_resources(model, payload, user, **kwargs)
            return self(model, user, **kwargs)
        except ValidationError as err:
            raise ApiError(merge_dict(ApiError.RESOURCE_INVALID, {
                # We need to return 200 so the front end client can process the error response
                # and show the user helpful errors.
                'code': 200,
                'errors': err.to_dict(),
            }))

    @classmethod
    @safe_db_query
    def create_associated_resources(self, model, payload, user, **kwargs):
        """
        Subclasses override this method
        """
        pass

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        try:
            db_connection.session.commit()
        except Exception:
            db_connection.session.rollback()

        model = self.model_class.query.get(pk)
        if not model:
            raise DoesNotExistError(f'{self.model_class.__name__} {pk} does not exist.')
        return self(model, user, **kwargs)

    @safe_db_query
    def delete(self, **kwargs):
        return self.model.delete()

    @safe_db_query
    def update(self, payload, **kwargs):
        for k, v in payload.items():
            setattr(self.model, k, v)
        self.update_associated_resources(payload, **kwargs)
        self.model.save()
        return self

    @safe_db_query
    def update_associated_resources(self, payload, **kwargs):
        """
        Subclasses override this method
        """
        pass



import os
from typing import Dict

from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.services.gcp.cloud_run import cloud_run
from mage_ai.shared.hash import merge_dict


class GcpCloudRunBlockExecutor(BlockExecutor):
    def __init__(self, pipeline, block_uuid: str, execution_partition: str = None):
        super().__init__(pipeline, block_uuid, execution_partition=execution_partition)
        self.executor_config = self.pipeline.repo_config.gcp_cloud_run_config or dict()
        if os.getenv('GCP_REGION'):
            self.executor_config['region'] = os.getenv('GCP_REGION')
        if os.getenv('GCP_PROJECT_ID'):
            self.executor_config['project_id'] = os.getenv('GCP_PROJECT_ID')
        if self.block.executor_config is not None:
            self.executor_config = merge_dict(self.executor_config, self.block.executor_config)

    def execute(
        self,
        block_run_id: int = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> None:
        cmd = self._run_commands(block_run_id, global_vars, **kwargs)
        cloud_run.run_job(
            ' '.join(cmd),
            f'mage-data-prep-block-{block_run_id}',
            cloud_run_config=self.executor_config,
        )



# from mage_ai.extensions.dbt.constants import (
#     EXTENSION_DESCRIPTION as EXTENSION_DESCRIPTION_GREAT_DBT,
#     EXTENSION_NAME as EXTENSION_NAME_GREAT_DBT,
#     EXTENSION_UUID as EXTENSION_UUID_GREAT_DBT,
# )
from mage_ai.extensions.great_expectations.constants import (
    EXTENSION_DESCRIPTION as EXTENSION_DESCRIPTION_GREAT_EXPECTATIONS,
    EXTENSION_NAME as EXTENSION_NAME_GREAT_EXPECTATIONS,
    EXTENSION_TEMPLATES as EXTENSION_TEMPLATES_GREAT_EXPECTATIONS,
    EXTENSION_UUID as EXTENSION_UUID_GREAT_EXPECTATIONS,
)
from mage_ai.shared.hash import index_by

EXTENSIONS = [
    dict(
        description=EXTENSION_DESCRIPTION_GREAT_EXPECTATIONS,
        name=EXTENSION_NAME_GREAT_EXPECTATIONS,
        templates=EXTENSION_TEMPLATES_GREAT_EXPECTATIONS,
        uuid=EXTENSION_UUID_GREAT_EXPECTATIONS,
    ),
    # dict(
    #     description=EXTENSION_DESCRIPTION_GREAT_DBT,
    #     name=EXTENSION_NAME_GREAT_DBT,
    #     uuid=EXTENSION_UUID_GREAT_DBT,
    # ),
]

EXTENSIONS_BY_UUID = index_by(lambda x: x['uuid'], EXTENSIONS)

EXTENSION_UUIDS = [d['uuid'] for d in EXTENSIONS]



from datetime import datetime
from mage_ai.api.errors import ApiError
from mage_ai.api.operations.constants import META_KEY_LIMIT
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import BlockRun, PipelineRun, PipelineSchedule
from sqlalchemy.orm import aliased
from typing import Dict, List

MAX_LOG_FILES = 20


class LogResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        parent_model = kwargs['parent_model']

        arr = []
        if type(parent_model) is BlockRun:
            arr = parent_model.logs
        elif issubclass(parent_model.__class__, Pipeline):
            arr = await self.__pipeline_logs(parent_model, query, meta)

        return self.build_result_set(
            arr,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    async def __pipeline_logs(self, pipeline: Pipeline, query_arg, meta) -> List[Dict]:
        pipeline_uuid = pipeline.uuid

        start_timestamp = query_arg.get('start_timestamp', [None])
        if start_timestamp:
            start_timestamp = start_timestamp[0]
        end_timestamp = query_arg.get('end_timestamp', [None])
        if end_timestamp:
            end_timestamp = end_timestamp[0]

        error = ApiError.RESOURCE_INVALID.copy()
        if start_timestamp:
            try:
                start_timestamp = datetime.fromtimestamp(int(start_timestamp))
            except (ValueError, OverflowError):
                error.update(message='Value is invalid for start_timestamp.')
                raise ApiError(error)
        if end_timestamp:
            try:
                end_timestamp = datetime.fromtimestamp(int(end_timestamp))
            except (ValueError, OverflowError):
                error.update(message='Value is invalid for end_timestamp.')
                raise ApiError(error)

        pipeline_schedule_ids = query_arg.get('pipeline_schedule_id[]', [None])
        if pipeline_schedule_ids:
            pipeline_schedule_ids = pipeline_schedule_ids[0]
        if pipeline_schedule_ids:
            pipeline_schedule_ids = pipeline_schedule_ids.split(',')
        else:
            pipeline_schedule_ids = []

        block_uuids = query_arg.get('block_uuid[]', [None])
        if block_uuids:
            block_uuids = block_uuids[0]
        if block_uuids:
            block_uuids = block_uuids.split(',')
        else:
            block_uuids = []

        pipeline_run_ids = query_arg.get('pipeline_run_id[]', [None])
        if pipeline_run_ids:
            pipeline_run_ids = pipeline_run_ids[0]
        if pipeline_run_ids:
            pipeline_run_ids = pipeline_run_ids.split(',')
        else:
            pipeline_run_ids = []

        block_run_ids = query_arg.get('block_run_id[]', [None])
        if block_run_ids:
            block_run_ids = block_run_ids[0]
        if block_run_ids:
            block_run_ids = block_run_ids.split(',')
        else:
            block_run_ids = []

        a = aliased(PipelineRun, name='a')
        b = aliased(PipelineSchedule, name='b')
        c = aliased(BlockRun, name='c')

        columns = [
            a.execution_date,
            a.pipeline_schedule_id,
            a.pipeline_schedule_id,
            a.pipeline_uuid,
        ]

        total_pipeline_run_log_count = 0
        pipeline_run_logs = []

        @safe_db_query
        def get_pipeline_runs():
            query = (
                PipelineRun.
                select(*columns).
                join(b, a.pipeline_schedule_id == b.id).
                filter(b.pipeline_uuid == pipeline_uuid).
                order_by(a.created_at.desc())
            )

            if len(pipeline_schedule_ids):
                query = (
                    query.
                    filter(a.pipeline_schedule_id.in_(pipeline_schedule_ids))
                )

            if len(pipeline_run_ids):
                query = (
                    query.
                    filter(a.id.in_(pipeline_run_ids))
                )

            if start_timestamp:
                query = (
                    query.
                    filter(a.execution_date >= start_timestamp)
                )

            if end_timestamp:
                query = (
                    query.
                    filter(a.execution_date <= end_timestamp)
                )
            total_pipeline_run_log_count = query.count()
            if meta.get(META_KEY_LIMIT, None) is not None:
                rows = query.limit(meta[META_KEY_LIMIT])
            else:
                rows = query.all()
            return dict(
                total_pipeline_run_log_count=total_pipeline_run_log_count,
                rows=rows,
            )

        if not len(block_uuids) and not len(block_run_ids):
            pipeline_run_results = get_pipeline_runs()
            total_pipeline_run_log_count = pipeline_run_results['total_pipeline_run_log_count']
            pipeline_run_rows = pipeline_run_results['rows']

            processed_pipeline_run_log_files = set()
            for row in pipeline_run_rows:
                model = PipelineRun()
                model.execution_date = row.execution_date
                model.pipeline_schedule_id = row.pipeline_schedule_id
                model.pipeline_uuid = row.pipeline_uuid
                logs = await model.logs_async()
                pipeline_log_file_path = logs.get('path')
                if pipeline_log_file_path not in processed_pipeline_run_log_files:
                    pipeline_run_logs.append(logs)
                    processed_pipeline_run_log_files.add(pipeline_log_file_path)
                if len(pipeline_run_logs) >= MAX_LOG_FILES:
                    break

        @safe_db_query
        def get_block_runs():
            query = (
                BlockRun.
                select(*(columns + [
                    c.block_uuid,
                ])).
                join(a, a.id == c.pipeline_run_id).
                join(b, a.pipeline_schedule_id == b.id).
                filter(b.pipeline_uuid == pipeline_uuid).
                order_by(c.started_at.desc())
            )

            if len(block_uuids):
                query = (
                    query.
                    filter(c.block_uuid.in_(block_uuids))
                )

            if len(block_run_ids):
                query = (
                    query.
                    filter(c.id.in_(block_run_ids))
                )

            if len(pipeline_schedule_ids):
                query = (
                    query.
                    filter(a.pipeline_schedule_id.in_(pipeline_schedule_ids))
                )

            if len(pipeline_run_ids):
                query = (
                    query.
                    filter(a.id.in_(pipeline_run_ids))
                )

            if start_timestamp:
                query = (
                    query.
                    filter(a.execution_date >= start_timestamp)
                )

            if end_timestamp:
                query = (
                    query.
                    filter(a.execution_date <= end_timestamp)
                )

            if meta.get(META_KEY_LIMIT, None) is not None:
                rows = query.limit(meta[META_KEY_LIMIT])
            else:
                rows = query.all()

            total_block_run_log_count = query.count()
            return dict(
                total_block_run_log_count=total_block_run_log_count,
                rows=rows,
            )

        block_run_results = get_block_runs()
        total_block_run_log_count = block_run_results['total_block_run_log_count']
        rows = block_run_results['rows']

        block_run_logs = []

        processed_block_run_log_files = set()
        for row in rows:
            model = PipelineRun()
            model.execution_date = row.execution_date
            model.pipeline_schedule_id = row.pipeline_schedule_id
            model.pipeline_uuid = row.pipeline_uuid

            model2 = BlockRun()
            model2.block_uuid = row.block_uuid
            model2.pipeline_run = model

            logs = await model2.logs_async()
            block_log_file_path = logs.get('path')
            if block_log_file_path not in processed_block_run_log_files:
                block_run_logs.append(logs)
                processed_block_run_log_files.add(block_log_file_path)

            if len(block_run_logs) >= MAX_LOG_FILES:
                break

        return [
            dict(
                block_run_logs=block_run_logs,
                pipeline_run_logs=pipeline_run_logs,
                total_block_run_log_count=total_block_run_log_count,
                total_pipeline_run_log_count=total_pipeline_run_log_count,
            ),
        ]



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.templates.constants import TEMPLATES, TEMPLATES_BY_UUID
from mage_ai.orchestration.db import safe_db_query


class BlockTemplateResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        return self.build_result_set(
            TEMPLATES,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        model = TEMPLATES_BY_UUID.get(pk)
        if not model:
            raise ApiError(ApiError.RESOURCE_NOT_FOUND)

        return self(model, user, **kwargs)



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.GitBranchPresenter import GitBranchPresenter


class GitBranchPolicy(BasePolicy):
    pass


GitBranchPolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

GitBranchPolicy.allow_actions([
    constants.CREATE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role())

GitBranchPolicy.allow_read(GitBranchPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE
], on_action=[
    constants.CREATE,
    constants.DETAIL,
    constants.LIST,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

GitBranchPolicy.allow_write(GitBranchPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role())

GitBranchPolicy.allow_write(GitBranchPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role())



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.server.api.base import BaseHandler
from mage_ai.server.constants import VERSION
import aiohttp


class ApiProjectsHandler(BaseHandler):
    async def get(self):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    'https://pypi.org/pypi/mage-ai/json',
                    timeout=3,
                ) as response:
                    response_json = await response.json()
                    latest_version = response_json.get('info', {}).get('version', None)
        except Exception:
            latest_version = VERSION
        parts = get_repo_path().split('/')
        collection = [
            dict(
                latest_version=latest_version,
                name=parts[-1],
                version=VERSION,
            ),
        ]
        self.write(dict(projects=collection))



from mage_ai.autocomplete.utils import extract_all_imports
from mage_ai.data_preparation.repo_manager import get_repo_path
import importlib
import re


def reload_all_repo_modules(content: str) -> None:
    parts = get_repo_path().split('/')
    project_name = parts[-1]

    for line in extract_all_imports(content):
        if f'import {project_name}' not in line and f'from {project_name}' not in line:
            continue

        regex = re.compile(r'from ([\w.]+) import|import ([\w.]+)')
        matches = re.findall(regex, line)

        if len(matches) >= 1:
            m1, m2 = matches[0]
            try:
                importlib.reload(importlib.import_module(m1 or m2))
            except Exception:
                pass



from mage_ai.data_cleaner.cleaning_rules.clean_column_names import CleanColumnNames
from mage_ai.data_cleaner.transformer_actions.constants import ActionType
from mage_ai.tests.base_test import TestCase
import pandas as pd


class CleanColumnNameTests(TestCase):
    def test_evaluate(self):
        df = pd.DataFrame(
            [
                ['', '', '', '', '', '', '', '', '', '', ''],
            ],
            columns=[
                'good_name',
                'Bad Case',
                '%@#342%34@@#342',
                'yield',
                '12342',
                '1234.    23',
                'true_crime',
                '@#f$%&*o$*(%^&r*$%&',
                'PascalCaseTitle',
                'camelCaseName',
                'lowercase9234',
            ],
        )
        result = CleanColumnNames(
            df,
            {},
            {},
        ).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Clean dirty column names',
                    message='Format these column names as lowercase and alphanumeric to improve '
                    'accessibility and security.',
                    status='not_applied',
                    action_payload=dict(
                        action_type=ActionType.CLEAN_COLUMN_NAME,
                        action_arguments=[
                            'Bad Case',
                            '%@#342%34@@#342',
                            'yield',
                            '12342',
                            '1234.    23',
                            '@#f$%&*o$*(%^&r*$%&',
                            'PascalCaseTitle',
                            'camelCaseName',
                        ],
                        action_code='',
                        action_options={},
                        action_variables={},
                        axis='column',
                        outputs=[],
                    ),
                )
            ],
        )



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import safe_db_query
from mage_ai.shared.array import find
import urllib.parse


class FileContentResource(GenericResource):
    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        file_path = urllib.parse.unquote(pk)
        file = File.from_path(file_path, get_repo_path())
        return self(file, user, **kwargs)

    def update(self, payload, **kwargs):
        error = ApiError.RESOURCE_INVALID.copy()

        content = payload.get('content')
        version = payload.get('version')

        if content is None and version is None:
            error.update(message='Please provide a content or version in the request payload.')
            raise ApiError(error)

        if version is not None and content is None:
            file_version = find(lambda f: f.filename == str(version), self.model.file_versions())
            content = file_version.content()

        self.model.update_content(content)



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
)


class RemoveDuplicateRows(BaseRule):
    def evaluate(self):
        df_dedupe = self.df.drop_duplicates()
        duplicate_row_count = self.df.shape[0] - df_dedupe.shape[0]
        suggestions = []
        if duplicate_row_count > 0:
            suggestions.append(
                self._build_transformer_action_suggestion(
                    'Remove duplicate rows',
                    f'Remove {duplicate_row_count} duplicated row(s) to reduce the amount of '
                    'redundant data.',
                    ActionType.DROP_DUPLICATE,
                    action_arguments=[],
                    axis=Axis.ROW,
                )
            )
        return suggestions



"""Add token column to pipeline_schedule table

Revision ID: 7ac6fed06918
Revises: ec5df57a1c60
Create Date: 2023-01-02 14:32:01.533214

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '7ac6fed06918'
down_revision = 'ec5df57a1c60'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('pipeline_schedule', sa.Column('token', sa.String(length=255), nullable=True))
    op.create_index(op.f('ix_pipeline_schedule_token'), 'pipeline_schedule', ['token'], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_pipeline_schedule_token'), table_name='pipeline_schedule')
    op.drop_column('pipeline_schedule', 'token')
    # ### end Alembic commands ###



from mage_ai.streaming.sources.rabbitmq import RabbitMQSource
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class PikaTests(TestCase):
    def test_init(self):
        with patch.object(RabbitMQSource, 'init_client') as mock_init_client:
            RabbitMQSource(dict(
                connection_host='localhost',
                connection_port='5672',
                queue_name='test_queue',
                configure_consume=False,
                username='guest',
                password='guest',
                amqp_url_virtual_host=r'%2f',
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(RabbitMQSource, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                RabbitMQSource(dict(
                    connection_port='5672',
                    queue_name='test_queue',
                    configure_consume=False,
                    username='guest',
                    password='guest',
                    amqp_url_virtual_host=r'%2f',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'connection_host\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db import safe_db_query
from mage_ai.server.scheduler_manager import scheduler_manager


class SchedulerResource(GenericResource):
    @classmethod
    def collection(self, query, meta, user, **kwargs):
        scheduler = dict(status=scheduler_manager.get_status())
        return self.build_result_set(
            [scheduler],
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        action_type = payload.get('action_type')

        if action_type == 'start':
            scheduler_manager.start_scheduler()
        elif action_type == 'stop':
            scheduler_manager.stop_scheduler()

        return self(dict(status=scheduler_manager.get_status()), user, **kwargs)



from mage_ai.data_preparation.repo_manager import get_variables_dir
from mage_ai.orchestration.constants import (
    DATABASE_CONNECTION_URL_ENV_VAR,
    DB_NAME,
    DB_PASS,
    DB_USER
)
from mage_ai.shared.environments import is_dev, is_test
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
import logging
import os
import sqlalchemy

DB_RETRY_COUNT = 2
TEST_DB = 'test.db'

db_connection_url = os.getenv(DATABASE_CONNECTION_URL_ENV_VAR)
db_kwargs = dict(
    connect_args={},
    pool_pre_ping=True,
)

if is_test():
    db_connection_url = f'sqlite:///{TEST_DB}'
elif not db_connection_url:
    # connect to K8s CloudSQL sidecar
    if os.getenv(DB_USER):
        db_user = os.getenv(DB_USER)
        db_pass = os.getenv(DB_PASS)
        db_name = os.getenv(DB_NAME)
        db_connection_url = f'postgresql+psycopg2://{db_user}:{db_pass}@127.0.0.1:5432/{db_name}'
    else:
        if is_test():
            db_connection_url = f'sqlite:///{TEST_DB}'
        elif os.path.exists('mage_ai/orchestration/db/'):
            # For local dev environment
            db_connection_url = 'sqlite:///mage_ai/orchestration/db/mage-ai.db'
        elif os.path.exists('mage-ai.db'):
            # For backward compatiblility
            db_connection_url = f'sqlite:///{get_variables_dir()}/mage-ai.db'
        else:
            # For new projects, create mage-ai.db in variables idr
            db_connection_url = f'sqlite:///{get_variables_dir()}/mage-ai.db'
        db_kwargs['connect_args']['check_same_thread'] = False

if db_connection_url.startswith('postgresql'):
    db_kwargs['pool_size'] = 50
    db_kwargs['connect_args']['options'] = '-c timezone=utc'

engine = create_engine(
    db_connection_url,
    **db_kwargs,
)
session_factory = sessionmaker(bind=engine)


class DBConnection:
    def __init__(self):
        self.session = None

    def start_session(self, force: bool = False):
        if self.session is not None and self.session.is_active and not force:
            return
        self.session = scoped_session(session_factory)

    def close_session(self):
        self.session.close()
        self.session = None


db_connection = DBConnection()


def safe_db_query(func):
    def func_with_rollback(*args, **kwargs):
        retry_count = 0
        while True:
            try:
                return func(*args, **kwargs)
            except (
                sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.PendingRollbackError,
                sqlalchemy.exc.InternalError,
            ) as e:
                db_connection.session.rollback()
                if retry_count >= DB_RETRY_COUNT:
                    raise e
                retry_count += 1
    return func_with_rollback


logging.basicConfig()

if is_dev():
    logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)



from alembic.config import Config
from alembic import command
from mage_ai.orchestration.db import db_connection_url
from mage_ai.shared.logger import LoggingLevel
import os


class DatabaseManager:
    def __init__(self):
        pass

    @property
    def script_location(self):
        pass

    def run_migrations(self, log_level: LoggingLevel = None):
        cur_dirpath = os.path.abspath(os.path.dirname(__file__))
        alembic_cfg = Config(os.path.join(cur_dirpath, 'alembic.ini'))
        alembic_cfg.set_main_option(
            'script_location',
            os.path.join(cur_dirpath, 'migrations'),
        )
        alembic_cfg.set_main_option('sqlalchemy.url', db_connection_url)
        if log_level is not None:
            alembic_cfg.set_section_option('logger_alembic', 'level', log_level)
        command.upgrade(alembic_cfg, 'head')


class SqliteDatabaseManager(DatabaseManager):
    pass


database_manager = SqliteDatabaseManager()



from mage_ai.io.postgres import Postgres
from mage_ai.io.export_utils import infer_dtypes
from mage_ai.tests.base_test import TestCase
import datetime
import faker
import numpy as np
import pandas as pd


class TypeConversionTests(TestCase):
    def setUp(self):
        np.random.seed(42)
        fake = faker.Faker()
        self.data = pd.DataFrame(
            [
                [
                    fake.name(),
                    fake.date_of_birth(),
                    fake.text(max_nb_chars=100),
                    np.random.randint(4, 10) / 2,
                    np.random.randn() * 100 + 20,
                    True if np.random.uniform() > 0.3 else False,
                    np.random.choice(['Mage', 'Data Cleaning', 'Magic', 'Awesome']),
                    np.random.randint(-100, 100),
                    np.random.randint(0, 3),
                    fake.date_this_century(),
                    fake.time_object(),
                    fake.date_time_this_century(),
                    fake.time_delta(datetime.timedelta(hours=3)),
                    np.random.randint(10000000000, 1000000000000),
                ]
                for _ in range(6)
            ],
            columns=[
                'Name',
                'Date of Birth',
                'Review',
                'Stars',
                'Profit',
                'Verified Purchase',
                'Tags',
                'Percentage Used',
                'Number of Snacks',
                'Watch Date',
                'Watch Time',
                'Watch Date And Time',
                'Elapsed Time',
                'Ticket ID',
            ],
        )
        self.dtypes = infer_dtypes(self.data)
        return super().setUp()

    def test_postgres_detection(self):
        expected_dtypes = [
            'text',
            'date',
            'text',
            'double precision',
            'double precision',
            'boolean',
            'text',
            'smallint',
            'smallint',
            'date',
            'time',
            'timestamp',
            'bigint',
            'bigint',
        ]
        psql = Postgres('test', 'test', 'test', 'test', 'test', True)
        for column, expected_dtype in zip(self.data.columns, expected_dtypes):
            dtype = self.dtypes[column]
            psql_type = psql.get_type(self.data[column], dtype)
            self.assertEqual(psql_type, expected_dtype)



from enum import Enum
from mage_ai.data_cleaner.shared.utils import (
    is_geo_dataframe,
    is_spark_dataframe,
)
from mage_ai.data_preparation.models.constants import (
    DATAFRAME_ANALYSIS_KEYS,
    DATAFRAME_SAMPLE_COUNT,
    DATAFRAME_SAMPLE_MAX_COLUMNS,
    VARIABLE_DIR,
)
from mage_ai.data_preparation.models.utils import (
    apply_transform_pandas,
    # dask_from_pandas,
    cast_column_types,
    deserialize_columns,
    serialize_columns,
)
from mage_ai.data_preparation.storage.base_storage import BaseStorage
from mage_ai.data_preparation.storage.local_storage import LocalStorage
from mage_ai.shared.parsers import sample_output
from mage_ai.shared.utils import clean_name
from pandas.api.types import is_object_dtype
from typing import Any, Dict, List
import json
import numpy as np
import os
import pandas as pd
import polars as pl
import traceback

DATAFRAME_COLUMN_TYPES_FILE = 'data_column_types.json'
DATAFRAME_PARQUET_FILE = 'data.parquet'
DATAFRAME_PARQUET_SAMPLE_FILE = 'sample_data.parquet'
DATAFRAME_CSV_FILE = 'data.csv'

JSON_FILE = 'data.json'
JSON_SAMPLE_FILE = 'sample_data.json'


class VariableType(str, Enum):
    DATAFRAME = 'dataframe'
    DATAFRAME_ANALYSIS = 'dataframe_analysis'
    GEO_DATAFRAME = 'geo_dataframe'
    POLARS_DATAFRAME = 'polars_dataframe'
    SPARK_DATAFRAME = 'spark_dataframe'


class Variable:
    def __init__(
        self,
        uuid: str,
        pipeline_path: str,
        block_uuid: str,
        partition: str = None,
        spark=None,
        storage: BaseStorage = LocalStorage(),
        variable_type: VariableType = None
    ) -> None:
        self.uuid = uuid
        self.storage = storage
        # if not self.storage.path_exists(pipeline_path):
        #     raise Exception(f'Pipeline path {pipeline_path} does not exist.')
        self.pipeline_path = pipeline_path
        self.block_uuid = block_uuid
        self.block_dir_name = clean_name(self.block_uuid)
        self.partition = partition
        self.variable_dir_path = os.path.join(
            pipeline_path,
            VARIABLE_DIR,
            partition or '',
            self.block_dir_name,
        )
        if not self.storage.path_exists(self.variable_dir_path):
            self.storage.makedirs(self.variable_dir_path)

        self.variable_type = variable_type
        self.check_variable_type(spark=spark)

    @property
    def variable_path(self):
        return os.path.join(self.variable_dir_path, f'{self.uuid}')

    @classmethod
    def dir_path(self, pipeline_path, block_uuid):
        return os.path.join(pipeline_path, VARIABLE_DIR, clean_name(block_uuid))

    def check_variable_type(self, spark=None):
        """
        Infer variable type based on data in the storage.
        """
        if self.variable_type is None and self.storage.path_exists(
            os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE)
        ):
            # If parquet file exists for given variable, set the variable type to DATAFRAME
            self.variable_type = VariableType.DATAFRAME
        elif ((self.variable_type == VariableType.DATAFRAME or self.variable_type is None)
                and os.path.exists(
                os.path.join(self.variable_dir_path, f'{self.uuid}', 'data.sh'))):
            self.variable_type = VariableType.GEO_DATAFRAME
        elif self.variable_type is None and \
                len(self.storage.listdir(self.variable_path, suffix='.parquet')) > 0 and \
                spark is not None:
            self.variable_type = VariableType.SPARK_DATAFRAME

    def convert_parquet_to_csv(self):
        """
        For DATAFRAME variable, convert parquet files to csv files.
        """
        if self.variable_type != VariableType.DATAFRAME:
            return
        csv_file_path = os.path.join(self.variable_path, DATAFRAME_CSV_FILE)
        if self.storage.path_exists(csv_file_path):
            return
        df = self.__read_parquet()
        self.storage.write_csv(df, csv_file_path)

    def delete(self):
        """
        Delete the variable data.
        """
        if self.variable_type is None and self.storage.path_exists(
            os.path.join(self.variable_dir_path, f'{self.uuid}', DATAFRAME_PARQUET_FILE)
        ):
            # If parquet file exists for given variable, set the variable type to DATAFRAME
            self.variable_type = VariableType.DATAFRAME
        if self.variable_type == VariableType.DATAFRAME:
            self.__delete_parquet()
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return self.__delete_dataframe_analysis()
        return self.__delete_json()

    def read_data(
        self,
        dataframe_analysis_keys: List[str] = None,
        sample: bool = False,
        sample_count: int = None,
        spark=None,
    ) -> Any:
        """
        Read variable data.

        Args:
            dataframe_analysis_keys (List[str], optional): For DATAFRAME_ANALYSIS variable,
                only read the selected keys.
            sample (bool, optional): Whether to sample the rows of a dataframe, used for
                DATAFRAME variable.
            sample_count (int, optional): The number of rows to sample, used for
                DATAFRAME variable.
            spark (None, optional): Spark context, used to read SPARK_DATAFRAME variable.
        """
        if self.variable_type == VariableType.DATAFRAME:
            return self.__read_parquet(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            return self.__read_spark_parquet(sample=sample, sample_count=sample_count, spark=spark)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            return self.__read_geo_dataframe(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return self.__read_dataframe_analysis(dataframe_analysis_keys=dataframe_analysis_keys)
        return self.__read_json(sample=sample)

    async def read_data_async(
        self,
        dataframe_analysis_keys: List[str] = None,
        sample: bool = False,
        sample_count: int = None,
        spark=None,
    ):
        """
        Read variable data asynchronously.

        Args:
            dataframe_analysis_keys (List[str], optional): For DATAFRAME_ANALYSIS variable,
                only read the selected keys.
            sample (bool, optional): Whether to sample the rows of a dataframe, used for
                DATAFRAME variable.
            sample_count (int, optional): The number of rows to sample, used for
                DATAFRAME variable.
            spark (None, optional): Spark context, used to read SPARK_DATAFRAME variable.
        """
        if self.variable_type == VariableType.DATAFRAME:
            return self.__read_parquet(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            return self.__read_spark_parquet(sample=sample, sample_count=sample_count, spark=spark)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return await self.__read_dataframe_analysis_async(
                dataframe_analysis_keys=dataframe_analysis_keys,
            )
        return await self.__read_json_async(sample=sample)

    def write_data(self, data: Any) -> None:
        """
        Write variable data to the persistent storage.

        Args:
            data (Any): Variable data to be written to storage.
        """
        if self.variable_type is None and type(data) is pd.DataFrame:
            self.variable_type = VariableType.DATAFRAME
        elif self.variable_type is None and type(data) is pl.DataFrame:
            self.variable_type = VariableType.POLARS_DATAFRAME
        elif is_spark_dataframe(data):
            self.variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            self.variable_type = VariableType.GEO_DATAFRAME

        if self.variable_type == VariableType.DATAFRAME:
            self.__write_parquet(data)
        elif self.variable_type == VariableType.POLARS_DATAFRAME:
            self.__write_polars_dataframe(data)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            self.__write_spark_parquet(data)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            self.__write_geo_dataframe(data)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            self.__write_dataframe_analysis(data)
        else:
            self.__write_json(data)

    async def write_data_async(self, data: Any) -> None:
        """
        Write variable data to the persistent storage.

        Args:
            data (Any): Variable data to be written to storage.
        """
        if self.variable_type is None and type(data) is pd.DataFrame:
            self.variable_type = VariableType.DATAFRAME
        elif self.variable_type is None and type(data) is pl.DataFrame:
            self.variable_type = VariableType.POLARS_DATAFRAME
        elif is_spark_dataframe(data):
            self.variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            self.variable_type = VariableType.GEO_DATAFRAME

        if self.variable_type == VariableType.DATAFRAME:
            self.__write_parquet(data)
        elif self.variable_type == VariableType.POLARS_DATAFRAME:
            self.__write_polars_dataframe(data)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            self.__write_spark_parquet(data)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            self.__write_geo_dataframe(data)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            self.__write_dataframe_analysis(data)
        else:
            await self.__write_json_async(data)

    def __delete_dataframe_analysis(self) -> None:
        for k in DATAFRAME_ANALYSIS_KEYS:
            file_path = os.path.join(self.variable_path, f'{k}.json')
            if self.storage.path_exists(file_path):
                try:
                    self.storage.remove(file_path)
                except FileNotFoundError as err:
                    print(f'Error deleting file {file_path}: {err}')

    def __delete_json(self) -> None:
        old_file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        if self.storage.path_exists(old_file_path):
            self.storage.remove(old_file_path)
        if self.storage.isdir(self.variable_path):
            self.storage.remove_dir(self.variable_path)

    def __delete_parquet(self) -> None:
        file_path = os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE)

        if self.storage.path_exists(file_path):
            self.storage.remove(file_path)
            self.storage.remove_dir(self.variable_path)

    def __read_json(self, default_value={}, sample: bool = False) -> Dict:
        # For backward compatibility
        old_file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)

        read_sample_success = False
        if sample and self.storage.path_exists(sample_file_path):
            try:
                data = self.storage.read_json_file(sample_file_path, default_value)
                read_sample_success = True
            except Exception:
                pass
        if not read_sample_success:
            if self.storage.path_exists(file_path):
                data = self.storage.read_json_file(file_path, default_value)
            else:
                data = self.storage.read_json_file(old_file_path, default_value)
        if sample:
            data = sample_output(data)[0]
        return data

    async def __read_json_async(self, default_value={}, sample: bool = False) -> Dict:
        # For backward compatibility
        old_file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)

        read_sample_success = False
        if sample and self.storage.path_exists(sample_file_path):
            try:
                data = await self.storage.read_json_file_async(sample_file_path, default_value)
                read_sample_success = True
            except Exception:
                pass
        if not read_sample_success:
            if self.storage.path_exists(file_path):
                data = await self.storage.read_json_file_async(file_path, default_value)
            else:
                data = await self.storage.read_json_file_async(old_file_path, default_value)
        if sample:
            data = sample_output(data)[0]
        return data

    def __write_json(self, data) -> None:
        if not self.storage.isdir(self.variable_path):
            self.storage.makedirs(self.variable_path, exist_ok=True)
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)
        self.storage.write_json_file(file_path, data)
        self.storage.write_json_file(sample_file_path, sample_output(data)[0])

    async def __write_json_async(self, data) -> None:
        if not self.storage.isdir(self.variable_path):
            self.storage.makedirs(self.variable_path, exist_ok=True)
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)
        try:
            await self.storage.write_json_file_async(file_path, data)
            await self.storage.write_json_file_async(sample_file_path, sample_output(data)[0])
        except Exception:
            traceback.print_exc()

    def __read_geo_dataframe(self, sample: bool = False, sample_count: int = None):
        import geopandas as gpd

        file_path = os.path.join(self.variable_path, 'data.sh')
        sample_file_path = os.path.join(self.variable_path, 'sample_data.sh')
        if not os.path.exists(file_path):
            return gpd.GeoDataFrame()
        if sample and os.path.exists(sample_file_path):
            try:
                df = gpd.read_file(sample_file_path)
            except Exception:
                df = gpd.read_file(file_path)
        else:
            df = gpd.read_file(file_path)
        if sample:
            sample_count = sample_count or DATAFRAME_SAMPLE_COUNT
            if df.shape[0] > sample_count:
                df = df.iloc[:sample_count]
        return df

    def __read_parquet(self, sample: bool = False, sample_count: int = None) -> pd.DataFrame:
        file_path = os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE)
        sample_file_path = os.path.join(self.variable_path, DATAFRAME_PARQUET_SAMPLE_FILE)

        read_sample_success = False
        if sample:
            try:
                df = self.storage.read_parquet(sample_file_path, engine='pyarrow')
                read_sample_success = True
            except Exception:
                pass
        if not read_sample_success:
            try:
                df = self.storage.read_parquet(file_path, engine='pyarrow')
            except Exception:
                df = pd.DataFrame()
        if sample:
            sample_count = sample_count or DATAFRAME_SAMPLE_COUNT
            if df.shape[0] > sample_count:
                df = df.iloc[:sample_count]

        column_types_filename = os.path.join(self.variable_path, DATAFRAME_COLUMN_TYPES_FILE)
        if os.path.exists(column_types_filename):
            with open(column_types_filename, 'r') as f:
                column_types = json.load(f)
                # ddf = dask_from_pandas(df)
                df = apply_transform_pandas(df, lambda row: deserialize_columns(row, column_types))
                df = cast_column_types(df, column_types)
        return df

    def __read_spark_parquet(self, sample: bool = False, sample_count: int = None, spark=None):
        if spark is None:
            return None
        df = (
            spark.read
            .format('parquet')
            .option('header', 'true')
            .option('inferSchema', 'true')
            .option('delimiter', ',')
            .load(self.variable_path)
        )
        if sample and sample_count:
            df = df.limit(sample_count)
        return df

    def __write_geo_dataframe(self, data) -> None:
        os.makedirs(self.variable_path, exist_ok=True)
        data.to_file(os.path.join(self.variable_path, 'data.sh'))
        df_sample_output = data.iloc[:DATAFRAME_SAMPLE_COUNT]
        df_sample_output.to_file(os.path.join(self.variable_path, 'sample_data.sh'))

    def __write_parquet(self, data: pd.DataFrame) -> None:
        column_types = {}
        df_output = data.copy()
        # Clean up data types since parquet doesn't support mixed data types
        for c in df_output.columns:
            c_dtype = df_output[c].dtype
            if not is_object_dtype(c_dtype):
                column_types[c] = str(c_dtype)
            else:
                series_non_null = df_output[c].dropna()
                if len(series_non_null) > 0:
                    coltype = type(series_non_null.iloc[0])

                    if is_object_dtype(series_non_null.dtype):
                        try:
                            df_output[c] = series_non_null.astype(coltype)
                        except Exception:
                            # Fall back to convert to string
                            # df_output[c] = series_non_null.astype(str)
                            pass

                    col_not_numpy_type = coltype.__module__ != np.__name__
                    col_is_numpy_array_type = coltype.__name__ == np.ndarray.__name__
                    if col_not_numpy_type or col_is_numpy_array_type:
                        column_types[c] = coltype.__name__
                    else:
                        column_types[c] = type(series_non_null.iloc[0].item()).__name__

        self.storage.makedirs(self.variable_path, exist_ok=True)

        # ddf = dask_from_pandas(df_output)
        df_output_serialized = apply_transform_pandas(
            df_output,
            lambda row: serialize_columns(row, column_types),
        )

        self.storage.write_parquet(
            df_output_serialized,
            os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE),
        )

        with open(os.path.join(self.variable_path, DATAFRAME_COLUMN_TYPES_FILE), 'w') as f:
            f.write(json.dumps(column_types))

        try:
            df_sample_output = df_output_serialized.iloc[
                :DATAFRAME_SAMPLE_COUNT,
                :DATAFRAME_SAMPLE_MAX_COLUMNS
            ]

            self.storage.write_parquet(
                df_sample_output,
                os.path.join(self.variable_path, DATAFRAME_PARQUET_SAMPLE_FILE),
            )
        except Exception as err:
            print(f'Sample output error: {err}.')
            traceback.print_exc()

    def __write_polars_dataframe(self, data: pl.DataFrame) -> None:
        self.storage.makedirs(self.variable_path, exist_ok=True)

        self.storage.write_polars_dataframe(
            data,
            os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE),
        )

        try:
            sample_columns = data.columns[:DATAFRAME_SAMPLE_MAX_COLUMNS]
            df_sample_output = data[
                :DATAFRAME_SAMPLE_COUNT,
                sample_columns,
            ]

            self.storage.write_polars_dataframe(
                df_sample_output,
                os.path.join(self.variable_path, DATAFRAME_PARQUET_SAMPLE_FILE),
            )
        except Exception as err:
            print(f'Sample output error: {err}.')
            traceback.print_exc()

    def __write_spark_parquet(self, data) -> None:
        (
            data.write
            .option('header', 'True')
            .mode('overwrite')
            .parquet(self.variable_path)
        )

    def __read_dataframe_analysis(
        self,
        dataframe_analysis_keys: List[str] = None,
    ) -> Dict[str, Dict]:
        """
        Read the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        if not self.storage.path_exists(self.variable_path):
            return dict()
        result = dict()
        for k in DATAFRAME_ANALYSIS_KEYS:
            if dataframe_analysis_keys is not None and k not in dataframe_analysis_keys:
                continue
            result[k] = self.storage.read_json_file(os.path.join(self.variable_path, f'{k}.json'))
        return result

    async def __read_dataframe_analysis_async(
        self,
        dataframe_analysis_keys: List[str] = None,
    ) -> Dict[str, Dict]:
        """
        Read the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        if not self.storage.path_exists(self.variable_path):
            return dict()
        result = dict()
        for k in DATAFRAME_ANALYSIS_KEYS:
            if dataframe_analysis_keys is not None and k not in dataframe_analysis_keys:
                continue
            result[k] = await self.storage.read_json_file_async(
                os.path.join(self.variable_path, f'{k}.json'))
        return result

    def __write_dataframe_analysis(self, data: Dict[str, Dict]) -> None:
        """
        Write the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        self.storage.makedirs(self.variable_path, exist_ok=True)
        for k in DATAFRAME_ANALYSIS_KEYS:
            self.storage.write_json_file(os.path.join(self.variable_path, f'{k}.json'), data.get(k))



from typing import Any


class FrozenDict(dict):
    def __setitem__(self, __k, __v) -> None:
        raise AttributeError('Cannot assign a value to frozen dictionary')

    def __delitem__(self, __v) -> None:
        raise AttributeError('Cannot delete a value from frozen dictionary')

    def clear(self) -> None:
        raise AttributeError('Cannot clear a frozen dictionary')

    def pop(self, __k) -> Any:
        raise AttributeError('Cannot remove item from a frozen dictionary')

    def popitem(self) -> Any:
        raise AttributeError('Cannot remove item from a frozen dictionary')

    def setdefault(self, __k, __v=None) -> Any:
        if __k in self.keys():
            return self[__k]
        else:
            raise AttributeError('Cannot assign a value to frozen dictionary')

    def update(self, other=None) -> Any:
        raise AttributeError('Cannot assign a value to frozen dictionary')

    def __hash__(self):
        # TODO: Come up with better hash function
        return hash(tuple((key, str(value)) for key, value in self.items()))



import re


def camel_to_snake_case(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    name = re.sub('__([A-Z])', r'_\1', name)
    name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name)
    return name.lower()


def format_enum(v):
    return v.value if type(v) is not str else v


def replacer(s, newstring, index, nofail=False):
    # raise an error if index is outside of the string
    if not nofail and index not in range(len(s)):
        raise ValueError('index outside given string')

    # if not erroring, but the index is still not in the correct range..
    if index < 0:  # add it to the beginning
        return newstring + s
    if index > len(s):  # add it to the end
        return s + newstring

    # insert the new string between 'slices' of the original
    return s[:index] + newstring + s[index + 1:]


def remove_extension_from_filename(filename: str) -> str:
    parts = filename.split('/')
    fn = parts[-1].split('.')[0]
    return '/'.join(parts[:-1] + [fn])



from datetime import datetime, timedelta
from mage_ai.services.dbt.config import DbtConfig
from mage_ai.services.dbt.constants import (
    DBT_CLOUD_BASE_URL,
    DEFAULT_POLL_INTERVAL,
    DbtCloudJobRunStatus,
)
from mage_ai.shared.http_client import HttpClient
from typing import Dict, List, Optional, Union
import time


class DbtCloudClient(HttpClient):
    """
    API doc: https://docs.getdbt.com/dbt-cloud/api-v2
    """

    BASE_URL = DBT_CLOUD_BASE_URL

    def __init__(self, config: Union[Dict, DbtConfig]):
        if type(config) is dict:
            self.config = DbtConfig.load(config=config)
        else:
            self.config = config
        self.headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Token {self.config.api_token}',
        }

    def list_jobs(
        self,
        order_by: str = None,
        project_id: str = None,
    ):
        """List jobs in a project
        API doc: https://docs.getdbt.com/dbt-cloud/api-v2#tag/Jobs

        Args:
            order_by (str, optional): Field to order the result by. Use - to indicate reverse order.
            project_id (str, optional): Numeric ID of the project containing jobs
        """
        return self.make_request(
            f'/{self.config.account_id}/jobs',
            params=dict(
                order_by=order_by,
                project_id=project_id,
            ),
        )

    def list_runs(
        self,
        include_related: List[str] = [],
        job_definition_id: int = None,
        order_by: str = None,
        offset: int = None,
        limit: int = 100,
    ):
        """List the runs for a given account:
        API doc: https://docs.getdbt.com/dbt-cloud/api-v2#tag/Runs/operation/listRunsForAccount

        Args:
            include_related (List[str], optional): List of related fields to pull with the run.
                Valid values are "trigger", "job", and "debug_logs". If "debug_logs" is not provided
                in a request, then the included debug logs will be truncated to the last 1,000 lines
                of the debug log output file.
            job_definition_id (int, optional): Applies a filter to only return runs from the
                specified Job.
            order_by (str, optional): Field to order the result by. Use - to indicate reverse order.
            offset (int, optional): The offset to apply when listing runs. Use with limit to
                paginate results.
            limit (int, optional): The limit to apply when listing runs. Use with offset to
                paginate results.
        """
        return self.make_request(
            f'/{self.config.account_id}/runs',
            params=dict(
                include_related=include_related,
                job_definition_id=job_definition_id,
                order_by=order_by,
                offset=offset,
                limit=limit,
            ),
        )

    def get_run(self, run_id: int, include_related: List[str] = []):
        return self.make_request(
            f'/{self.config.account_id}/runs/{run_id}',
            params=dict(
                include_related=include_related,
            ),
        )

    def trigger_job_run(
        self,
        job_id: int,
        cause: str = 'Trigger job run from Mage',
        git_sha: str = None,
        git_branch: str = None,
        schema_override: str = None,
        dbt_version_override: str = None,
        threads_override: int = None,
        target_name_override: str = None,
        generate_docs_override: bool = None,
        timeout_seconds_override: int = None,
        steps_override: List[str] = None,
        poll_status: bool = True,
        poll_interval: float = DEFAULT_POLL_INTERVAL,
        poll_timeout: Optional[float] = None,
    ):
        """Kick off a run for a job. When this endpoint returns a successful response,
        a new run will be enqueued for the account. Users can poll the Get run endpoint to
        poll the run until it completes. After the run has completed, users can use the Get
        run artifact endpoint to download artifacts generated by the run.
        API doc: https://docs.getdbt.com/dbt-cloud/api-v2#tag/Jobs/operation/triggerRun

        Args:
            job_id (int): Numeric ID of the job
            cause (str, optional): A text description of the reason for running this job
            git_sha (str, optional): The git sha to check out before running this job
            git_branch (str, optional): The git branch to check out before running this job
            schema_override (str, optional): Override the destination schema in the configured
                target for this job.
            dbt_version_override (str, optional): Override the version of dbt used to run this job
            threads_override (int, optional): Override the number of threads used to run this job
            target_name_override (str, optional): Override the target.name context variable used
                when running this job
            generate_docs_override (bool, optional): Override whether or not this job generates
                docs (true=yes, false=no)
            timeout_seconds_override (int, optional): Override the timeout in seconds for this job
            steps_override (List[str], optional): Override the list of steps for this job
            poll_status (bool, optional): Whether to poll stauts of the job run
            poll_interval (float, optional): Poll interval in seconds
            poll_timeout (Optional[float], optional): Poll timeout in seconds
        """
        job_run_response = self.make_request(
            f'/{self.config.account_id}/jobs/{job_id}/run',
            method='POST',
            payload=dict(
                cause=cause,
                git_sha=git_sha,
                schema_override=schema_override,
                dbt_version_override=dbt_version_override,
                threads_override=threads_override,
                target_name_override=target_name_override,
                generate_docs_override=generate_docs_override,
                timeout_seconds_override=timeout_seconds_override,
                steps_override=steps_override
            ),
        )
        if poll_status:
            run_id = job_run_response['data']['id']
            poll_start = datetime.now()
            while True:
                run_data = self.get_run(run_id)['data']
                job_run_status = run_data['status']
                job_run_status_msg = run_data['status_humanized']
                print(f'Polling DBT Cloud run {run_id}. Current status: {job_run_status_msg}.')

                if job_run_status in DbtCloudJobRunStatus.TERMINAL_STATUSES.value:
                    print(f"Job run status for run {run_id}: {job_run_status_msg}. "
                          "Polling complete")

                    if job_run_status == DbtCloudJobRunStatus.SUCCESS.value:
                        break
                    raise Exception(f'Job run {run_id} failed with status: {job_run_status_msg}.')
                if (
                    poll_timeout
                    and datetime.now()
                    > poll_start + timedelta(seconds=poll_timeout)
                ):
                    raise Exception(
                        f"Job run {run_id} for job {job_id} time out after "
                        f"{datetime.now() - poll_start}. Last status was {job_run_status_msg}."
                    )
                time.sleep(poll_interval)



from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.shared.utils import parse_list
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, VariableType
import numpy as np

DAY_SECONDS = 86400
HOUR_SECONDS = 3600


def convert_col_type(df_col, col_type):
    if col_type == ColumnType.NUMBER:
        return df_col.replace(r'^\s*$', 0, regex=True).fillna(0).astype(np.int64)
    elif col_type == ColumnType.NUMBER_WITH_DECIMALS:
        return df_col.dropna().astype(float)
    elif col_type == ColumnType.TEXT:
        return df_col.dropna().astype(str)
    elif col_type == ColumnType.LIST:
        return df_col.apply(parse_list)
    return df_col


def convert_value_type(feature_uuid, action, value):
    action_variables = action.get('action_variables', {})
    column_type = None
    for v in action_variables.values():
        if v['type'] == 'feature' and v['feature']['uuid'] == feature_uuid:
            column_type = v['feature']['column_type']
            break
    if column_type == ColumnType.NUMBER:
        value = int(value)
    elif column_type == ColumnType.NUMBER_WITH_DECIMALS:
        value = float(value)
    return value


def drop_na(df):
    return df.replace(r'^\s*$', np.nan, regex=True).dropna()


def extract_join_feature_set_version_id(payload):
    if payload['action_type'] != ActionType.JOIN:
        return None
    join_feature_set_version_id = payload['action_arguments'][0]
    if type(join_feature_set_version_id) == str and join_feature_set_version_id.startswith('%{'):

        join_feature_set_version_id = next(
            v['id']
            for v in payload['action_variables'].values()
            if v['type'] == VariableType.FEATURE_SET_VERSION
        )
    return join_feature_set_version_id


def get_column_type(feature_uuid, action):
    action_variables = action.get('action_variables', {})
    column_type = None
    for v in action_variables.values():
        if v['type'] == 'feature' and v['feature']['uuid'] == feature_uuid:
            column_type = v['feature']['column_type']
            break
    return column_type


def get_time_window_str(window_in_seconds):
    if window_in_seconds is None:
        return None
    if window_in_seconds >= DAY_SECONDS:
        time_window = f'{int(window_in_seconds / DAY_SECONDS)}d'
    elif window_in_seconds >= HOUR_SECONDS:
        time_window = f'{int(window_in_seconds / HOUR_SECONDS)}h'
    else:
        time_window = f'{window_in_seconds}s'
    return time_window
